import sys
import traceback
from abc import abstractmethod, ABC
from base64 import b64decode
from logging import Logger
from typing import Any

from flask import request, Response, Request
from sapiopylib.rest.DataRecordManagerService import DataRecordManager
from sapiopylib.rest.User import SapioUser
from sapiopylib.rest.WebhookService import AbstractWebhookHandler
from sapiopylib.rest.utils.DataTypeCacheManager import DataTypeCacheManager
from sapiopylib.rest.utils.recordmodel.RecordModelManager import RecordModelManager, RecordModelInstanceManager, \
    RecordModelRelationshipManager
from sapiopylib.rest.utils.recordmodel.ancestry import RecordModelAncestorManager
from werkzeug.datastructures import Headers
from werkzeug.datastructures.structures import MultiDict
from werkzeug.exceptions import UnsupportedMediaType, BadRequest

from sapiopycommons.general.exceptions import SapioException
from sapiopycommons.recordmodel.record_handler import RecordHandler


class SapioWebserviceException(Exception):
    """
    An exception to be thrown by webservice classes when responding to a request with an error.
    """
    msg: str
    code: int

    def __init__(self, msg: str, code: int = 500):
        """
        :param msg: The message to return in the webservice response.
        :param code: The status code to return in the webservice response.
        """
        self.msg = msg
        self.code = code


class SapioWebserviceResult:
    """
    A result to be returned by AbstractWebserviceHandler endpoints, for sending information back to the caller.
    """
    message: str
    is_error: bool
    status_code: int
    payload: dict[str, Any]

    def __init__(self, message: str = "Success", status_code: int = 200, is_error: bool = False,
                 payload: dict[str, Any] | None = None):
        """
        :param message: A message to return to the sender describing what happened.
        :param status_code: An HTTP status code to return to the sender.
        :param is_error: Whether the webservice had an error during processing.
        :param payload: A payload of additional information to return to the sender.
        """
        self.message = message
        self.status_code = status_code
        self.is_error = is_error
        self.payload = payload
        if payload is None:
            self.payload = {}

    def to_json(self) -> dict[str, Any]:
        return {"message": self.message,
                "statusCode": self.status_code,
                "isError": self.is_error,
                "payload": self.payload}

    def to_result(self) -> tuple[dict[str, Any], int]:
        return self.to_json(), self.status_code


WebserviceResponse = SapioWebserviceResult | Response | tuple[dict[str, Any] | int]


class AbstractWebserviceHandler(AbstractWebhookHandler):
    """
    A base class for constructing POST webservice endpoints on your webhook server. These are endpoints that can be
    communicated with by external sources without needing to format the request in the webhook context format that
    normal webhook handlers expect.

    Since this extends AbstractWebhookHandler, you can still register endpoints from this class in the same way you
    would normal webhook endpoints.
    """
    request: Request

    def post(self) -> Response | tuple[dict[str, Any], int]:
        """
        Internal method to be executed to translate incoming POST requests.
        """
        # noinspection PyBroadException
        try:
            self.request = request
            try:
                raw_json = request.json
                headers = request.headers
                params = request.args
            except UnsupportedMediaType as e:
                return SapioWebserviceResult(str(e), 415, True).to_json(), 415
            except BadRequest as e:
                return SapioWebserviceResult(str(e), 400, True).to_json(), 400
            ret_val: WebserviceResponse = self.run(raw_json, headers, params)
            if isinstance(ret_val, SapioWebserviceResult):
                return ret_val.to_result()
            return ret_val
        except Exception:
            print('Error occurred while running webservice custom logic. See traceback.', file=sys.stderr)
            traceback.print_exc()
            return SapioWebserviceResult("Unexpected error occurred.", 500, True).to_json(), 500

    # noinspection PyMethodOverriding
    @abstractmethod
    def run(self, payload: Any, headers: Headers, params: MultiDict[str, str]) -> WebserviceResponse:
        """
        The execution details for this endpoint.

        :param payload: The JSON payload from the request. Usually a dictionary of strings to Any.
        :param headers: The headers from the request. Can be used like a dictionary.
        :param params: The URL parameters from the request.
        :return: A response object to send back to the requester.
        """
        pass

    def authenticate_user(self, headers: dict[str, str]) -> SapioUser:
        """
        Authenticate a user for making requests to a Sapio server using the provided headers. If no user can be
        authenticated, then an exception will be thrown.

        :param headers: The headers of the webservice request.
        :return: A SapioUser object used to make requests to a Sapio server as authorized by the headers.
        """
        # Get the system URL from the headers.
        if "System-URL" not in headers:
            raise SapioWebserviceException("No \"System-URL\" provided in headers.", 400)
        url: str = headers.get("System-URL")
        if not url.endswith("/webservice/api"):
            raise SapioWebserviceException(f"\"System-URL\" must be a webservice API URL for the target system: {url}", 400)

        # Get the login credentials from the headers.
        auth: str = headers.get("Authorization")
        if auth and auth.startswith("Basic "):
            credentials: list[str] = b64decode(auth.split("Basic ")[1]).decode().split(":")
            user = self.basic_auth(url, credentials[0], credentials[1])
        elif auth and auth.startswith("Bearer "):
            user = self.bearer_token_auth(url, auth.split("Bearer ")[1])
        elif "X-API-TOKEN" in headers:
            user = self.api_token_auth(url, headers.get("X-API-TOKEN"))
        else:
            raise SapioWebserviceException(f"Unrecognized Authorization method.", 400)
        # Make a simple webservice call to confirm that the credentials are valid.
        try:
            # noinspection PyStatementEffect
            user.session_additional_data
        except Exception:
            if "Unauthorized (javax.ws.rs.NotAuthorizedException: Incorrect username or password.)" in traceback.format_exc():
                raise SapioWebserviceException("Unauthorized. Incorrect username or password.", 401)
            else:
                raise SapioWebserviceException("System-URL is invalid or user cannot be authenticated.", 401)
        return user

    def basic_auth(self, url: str, username: str, password: str) -> SapioUser:
        """
        :param url: The URL of the Sapio system that requests from this user will be sent to.
            Must end in /webservice/api
        :param username: The username to authenticate requests with.
        :param password: The password to authenticate requests with.
        :return: A SapioUser that will authenticate requests using basic auth.
        """
        return SapioUser(url, self.verify_sapio_cert, self.client_timeout_seconds, username=username, password=password)

    def api_token_auth(self, url: str, api_token: str) -> SapioUser:
        """
        :param url: The URL of the Sapio system that requests from this user will be sent to.
            Must end in /webservice/api
        :param api_token: The API token to authenticate requests with.
        :return: A SapioUser that will authenticate requests using an API token.
        """
        return SapioUser(url, self.verify_sapio_cert, self.client_timeout_seconds, api_token=api_token)

    def bearer_token_auth(self, url: str, bearer_token: str) -> SapioUser:
        """
        :param url: The URL of the Sapio system that requests from this user will be sent to.
            Must end in /webservice/api
        :param bearer_token: The bearer token to authenticate requests with.
        :return: A SapioUser that will authenticate requests using a bearer token.
        """
        return SapioUser(url, self.verify_sapio_cert, self.client_timeout_seconds, bearer_token=bearer_token)


class CommonsWebserviceHandler(AbstractWebserviceHandler, ABC):
    """
    A subclass of AbstractWebservicePostHandler that provides additional quality of life features, including
    authentication of a SapioUser from the request headers, initialization of various commonly used managers, and more.
    """
    logger: Logger

    user: SapioUser | None

    dr_man: DataRecordManager
    rec_man: RecordModelManager
    inst_man: RecordModelInstanceManager
    rel_man: RecordModelRelationshipManager
    an_man: RecordModelAncestorManager

    dt_cache: DataTypeCacheManager
    rec_handler: RecordHandler

    def run(self, payload: Any, headers: Headers, params: MultiDict[str, str]) -> WebserviceResponse:
        try:
            self.initialize(headers)
            result = self.execute(self.user, payload, headers, params)
            if result is None:
                raise SapioException("Your execute function returned a None result! Don't forget your return statement!")
            return result
        except SapioWebserviceException as e:
            return self.handle_webservice_exception(e)
        except Exception as e:
            return self.handle_unexpected_exception(e)

    def initialize(self, headers: Headers) -> None:
        """
        A function that can be optionally overridden by your classes to initialize additional instance variables,
        or set up whatever else you wish to set up before the execute function is ran. Default behavior initializes a
        SapioUser and various manager classes to make requests from.
        """
        self.user = None
        self.user = self.authenticate_user(headers)

        self.logger = self.user.logger

        self.dr_man = DataRecordManager(self.user)
        self.rec_man = RecordModelManager(self.user)
        self.inst_man = self.rec_man.instance_manager
        self.rel_man = self.rec_man.relationship_manager
        self.an_man = RecordModelAncestorManager(self.rec_man)

        self.dt_cache = DataTypeCacheManager(self.user)
        self.rec_handler = RecordHandler(self.user)

    @abstractmethod
    def execute(self, user: SapioUser, payload: Any, headers: Headers, params: MultiDict[str, str]) \
            -> SapioWebserviceResult:
        """
        The execution details for this endpoint.

        :param user: The SapioUser object authenticated from the request headers.
        :param payload: The JSON payload from the request. Usually a dictionary of strings to Any.
        :param headers: The headers from the request. Can be used like a dictionary.
        :param params: The URL parameters from the request.
        :return: A response object to send back to the requester.
        """
        pass

    def handle_webservice_exception(self, e: SapioWebserviceException) -> WebserviceResponse:
        """
        Handle a generic exception which isn't one of the handled Sapio exceptions.

        Default behavior returns a webservice result with the message and error code of the webservice exception.
        Additionally, the stack trace of the exception that was thrown is logged to the webhook server.

        :param e: The exception that was raised.
        :return: A webservice result to return to the requester.
        """
        result: WebserviceResponse | None = self.handle_any_exception(e)
        if result is not None:
            return result
        msg: str = traceback.format_exc()
        self.log_error(msg)
        return SapioWebserviceResult(e.msg, e.code, True)

    def handle_unexpected_exception(self, e: Exception) -> WebserviceResponse:
        """
        Handle a generic exception which isn't one of the handled Sapio exceptions.

        Default behavior returns a 500 code result with a generic error message informing the user to contact Sapio
        support. Additionally, the stack trace of the exception that was thrown is logged to the webhook server.

        :param e: The exception that was raised.
        :return: A webservice result to return to the requester.
        """
        result: SapioWebserviceResult | None = self.handle_any_exception(e)
        if result is not None:
            return result
        msg: str = traceback.format_exc()
        self.log_error(msg)
        return SapioWebserviceResult("Unexpected error occurred during webservice execution. "
                                     "Please contact Sapio support.", 500, True)

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def handle_any_exception(self, e: Exception) -> WebserviceResponse | None:
        """
        An exception handler which runs regardless of the type of exception that was raised. Can be used to "rollback"
        the client if an error occurs. Default behavior does nothing and returns None.

        :param e: The exception that was raised.
        :return: An optional response to the caller. May return a custom message to the client that wouldn't have been
            sent by one of the normal exception handlers, or may return None if no result needs returned. If a result is
            returned, then the default behavior of other exception handlers is skipped.
        """
        return None

    def log_info(self, msg: str) -> None:
        """
        Write an info message to the webhook server log. Log destination is stdout.
        """
        # If there's no user, then there's no logger.
        if self.user:
            self.logger.info(msg)

    def log_error(self, msg: str) -> None:
        """
        Write an error message to the webhook server log. Log destination is stderr.
        """
        # If there's no user, then there's no logger.
        if self.user:
            self.logger.error(msg)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any
from weakref import WeakValueDictionary

from sapiopylib.rest.User import SapioUser

_STR_JAVA_TYPE = "java.lang.String"
_INT_JAVA_TYPE = "java.lang.Integer"
_BOOL_JAVA_TYPE = "java.lang.Boolean"


class AbstractAccessionServiceOperator(ABC):
    """
    Abstract class to define an accession service operator.
    The default one in sapiopycommon only includes the out of box operators.
    More can be added via java plugins for global operators.
    """

    @property
    @abstractmethod
    def op_class_name(self) -> str:
        pass

    @property
    @abstractmethod
    def op_param_value_list(self) -> list[Any] | None:
        pass

    @property
    @abstractmethod
    def op_param_class_name_list(self) -> list[str] | None:
        pass

    @property
    @abstractmethod
    def default_accessor_name(self) -> str:
        pass


class AccessionWithPrefixSuffix(AbstractAccessionServiceOperator):
    """
    Local operator for accessioning prefix and suffix format.
    """
    _prefix: str | None
    _suffix: str | None
    _num_of_digits: int | None
    _start_num: int
    _strict_mode: bool

    @property
    def prefix(self):
        return self._prefix

    @property
    def suffix(self):
        return self._suffix

    @property
    def num_of_digits(self):
        return self._num_of_digits

    @property
    def start_num(self):
        return self._start_num

    @property
    def strict_mode(self):
        return self._strict_mode

    def __init__(self, prefix: str | None, suffix: str | None, num_of_digits: int | None = None,
                 start_num: int = 1, strict_mode: bool = False):
        if prefix is None:
            prefix = ""
        if suffix is None:
            suffix = ""
        self._prefix = prefix
        self._suffix = suffix
        self._num_of_digits = num_of_digits
        self._start_num = start_num
        self._strict_mode = strict_mode

    @property
    def op_param_value_list(self):
        return [self._prefix, self._suffix, self._num_of_digits, self._start_num, self._strict_mode]

    @property
    def op_param_class_name_list(self):
        return [_STR_JAVA_TYPE, _STR_JAVA_TYPE, _INT_JAVA_TYPE, _INT_JAVA_TYPE, _BOOL_JAVA_TYPE]

    @property
    def op_class_name(self):
        return "com.velox.accessionservice.operators.AccessionWithPrefixSuffix"

    @property
    def default_accessor_name(self):
        return "PREFIX_AND_SUFFIX" + "(" + self.prefix + "," + self.suffix + ")";


class AccessionGlobalPrefixSuffix(AbstractAccessionServiceOperator):
    """
    Global operator for accessioning prefix and suffix format.
    """
    _prefix: str | None
    _suffix: str | None
    _num_of_digits: int | None
    _start_num: int
    _strict_mode: bool

    @property
    def prefix(self):
        return self._prefix

    @property
    def suffix(self):
        return self._suffix

    @property
    def num_of_digits(self):
        return self._num_of_digits

    @property
    def start_num(self):
        return self._start_num

    @property
    def strict_mode(self):
        return self._strict_mode

    def __init__(self, prefix: str | None, suffix: str | None, num_of_digits: int | None = None,
                 start_num: int = 1, strict_mode: bool = False):
        if prefix is None:
            prefix = ""
        if suffix is None:
            suffix = ""
        self._prefix = prefix
        self._suffix = suffix
        self._num_of_digits = num_of_digits
        self._start_num = start_num
        self._strict_mode = strict_mode

    @property
    def op_param_value_list(self):
        return [self._prefix, self._suffix, self._num_of_digits, self._start_num, self._strict_mode]

    @property
    def op_param_class_name_list(self):
        return [_STR_JAVA_TYPE, _STR_JAVA_TYPE, _INT_JAVA_TYPE, _INT_JAVA_TYPE, _BOOL_JAVA_TYPE]

    @property
    def op_class_name(self):
        return "com.velox.accessionservice.operators.sapio.AccessionGlobalPrefixSuffix"

    @property
    def default_accessor_name(self):
        return "PREFIX_AND_SUFFIX" + "(" + self._prefix + "," + self._suffix + ")"


class AccessionNextBarcode(AbstractAccessionServiceOperator):
    """
    From Java description:
    This will start accessioning at the getNextBarcode() when there's no system preference to be backward compatible.
    However, once it completes setting the first ID, it will start increment by its own preference and disregards getNextBarcode().

    Recommend using AccessionServiceBasicManager to accession next barcode.
    To avoid ambiguity in preference cache.

    This should not be used unless we are using something legacy such as plate mapping template record creation
    (Note: not 3D plating, I'm talking about the older aliquoter).
    """

    @property
    def op_param_value_list(self):
        return []

    @property
    def op_param_class_name_list(self):
        return []

    @property
    def op_class_name(self):
        return "com.velox.accessionservice.operators.sapio.AccessionNextBarcode"

    @property
    def default_accessor_name(self):
        return "Barcode"


class AccessionRequestId(AbstractAccessionServiceOperator):
    """
    This class implements the accessioning operator for com.velox.sapioutils.shared.managers.DataRecordUtilManager.getNextRequestId()
    and getNextRequestId(int numberOfCharacters).

    Operation: For 4 characters start with A001, increment by 1 until A999. Then We use B001.
    After Z999 we start with AA01 until we get to AA99, etc.

    Exception: Skips I and O to prevent confusions with 1 and 0 when incrementing letters.

    Properties:
        numberOfCharacters: Number of characters maximum in the request ID.
        accessorName: This is a legacy variable from drum.getNextIdListByMapName(), which allows setting different "accessorName" from old system. We need this for compability patch for converting these to the new preference format.
    """
    _num_of_characters: int
    _accessor_name: str

    @property
    def num_of_characters(self):
        return self._num_of_characters

    @property
    def accessor_name(self):
        return self._accessor_name

    def __init__(self, num_of_characters: int = 4, accessor_name: str = None):
        self._num_of_characters = num_of_characters
        if not accessor_name:
            accessor_name = self.default_accessor_name
        self._accessor_name = accessor_name

    @property
    def op_class_name(self):
        return "com.velox.accessionservice.operators.sapio.AccessionRequestId"

    @property
    def op_param_value_list(self):
        return [self._num_of_characters, self._accessor_name]

    @property
    def op_param_class_name_list(self):
        return [_INT_JAVA_TYPE, _STR_JAVA_TYPE]

    @property
    def default_accessor_name(self):
        return "SapioNextRequestIdMap"


class AccessionServiceDescriptor:
    """
    Describes a single accession service's accessioning request

    Attributes:
        opClassName: The accession service operator class name as in Java
        opParamValueList: Ordered list of parameter values to construct the accession service operator.
        opParamClassNameList: Ordered list of FQCN of java classes in order of parameter value list.
        dataTypeName: The data type to accession. Should be blank if opClassName resolves to a global operator.
        dataFieldName: The data field to accession. Should be blank if opClassName resolves to a global operator.
        accessorName: The accessor cache name to be used for accessioning.
        numIds: The number of IDs to accession.
    """
    op: AbstractAccessionServiceOperator
    dataTypeName: str | None
    dataFieldName: str | None
    accessorName: str
    numIds: int

    def __init__(self, accessor_name: str, op: AbstractAccessionServiceOperator, num_ids: int,
                 data_type_name: str | None, data_field_name: str | None):
        self.accessorName = accessor_name
        self.op = op
        self.dataTypeName = data_type_name
        self.dataFieldName = data_field_name
        self.numIds = num_ids

    def to_json(self):
        return {
            "opClassName": self.op.op_class_name,
            "opParamValueList": self.op.op_param_value_list,
            "opParamClassNameList": self.op.op_param_class_name_list,
            "accessorName": self.accessorName,
            "numIds": self.numIds,
            "dataTypeName": self.dataTypeName,
            "dataFieldName": self.dataFieldName
        }


class AccessionService:
    """
    Provides Sapio Foundations Accession Service functionalities.
    """
    _user: SapioUser

    __instances: WeakValueDictionary[SapioUser, AccessionService] = WeakValueDictionary()
    __initialized: bool

    @property
    def user(self) -> SapioUser:
        return self._user

    def __new__(cls, user: SapioUser):
        """
        Observes singleton pattern per record model manager object.

        :param user: The user that will make the webservice request to the application.
        """
        obj = cls.__instances.get(user)
        if not obj:
            obj = object.__new__(cls)
            obj.__initialized = False
            cls.__instances[user] = obj
        return obj

    def __init__(self, user: SapioUser):
        if self.__initialized:
            return
        self._user = user
        self.__initialized = True

    def accession_with_config(self, data_type_name: str, data_field_name: str, num_ids: int) -> list[str]:
        """
        Accession with Configuration Manager => Accession Service configuration (This is not visible to regular users in SaaS)
        """
        payload = {
            "dataTypeName": data_type_name,
            "dataFieldName": data_field_name,
            "numIds": num_ids
        }
        response = self.user.plugin_post("accessionservice/accession_with_config", payload=payload)
        self.user.raise_for_status(response)
        return list(response.json())

    def accession_in_batch(self, descriptor: AccessionServiceDescriptor) -> list[str]:
        """
        This is the most flexible way to make use of accession service: directly via a descriptor object.
        """
        payload = descriptor.to_json()
        response = self.user.plugin_post("accessionservice/accession", payload=payload)
        self.user.raise_for_status(response)
        return list(response.json())

    def accession_next_request_id_list(self, num_of_characters: int, num_ids: int) -> list[str]:
        """
        Accession Request ID by old LIMS format. This is usually deprecated today.
        :param num_of_characters: Number of characters minimum in request ID.
        :param num_ids: Number of request IDs to accession.
        """
        op = AccessionRequestId(num_of_characters)
        descriptor = AccessionServiceDescriptor(op.default_accessor_name, op, num_ids, None, None)
        return self.accession_in_batch(descriptor)

    def get_affixed_id_in_batch(self, data_type_name: str, data_field_name: str, num_ids: int, prefix: str | None,
                                suffix: str | None, num_digits: int | None, start_num: int = 1) -> list[str]:
        """
        Get the batch affixed IDs that are maximal in cache and contiguious for a particular datatype.datafield under a given format.
        :param data_type_name: The datatype name to look for max ID
        :param data_field_name: The datafield name to look for max ID
        :param num_ids: The number of IDs to accession.
        :param prefix: leave it empty string "" if no prefix. Otherwise, specifies the prefix of ID.
        :param suffix: leave it empty string "" if no suffix. Otherwise, specifies the suffix of ID.
        :param num_digits: None if unlimited with no leading zeros.
        :param start_num The number to begin accessioning if this is the first time.
        :return:
        """
        op = AccessionWithPrefixSuffix(prefix, suffix, num_digits, start_num)
        descriptor = AccessionServiceDescriptor(op.default_accessor_name, op, num_ids, data_type_name, data_field_name)
        return self.accession_in_batch(descriptor)

    def get_global_affixed_id_in_batch(
            self, num_ids: int, prefix: str | None, suffix: str | None, num_digits: int | None, start_num: int = 1) -> list[str]:
        """
        Get the next numOfIds affixed IDs using system preference cache that's maximum across all datatype and datafields and maximal for the format.
        This method allows users to customize a start number instead of always starting at 1.
        :param num_ids: The number of IDs to accession.
        :param prefix: leave it empty string "" if no prefix. Otherwise, specifies the prefix of ID.
        :param suffix: leave it empty string "" if no suffix. Otherwise, specifies the suffix of ID.
        :param num_digits: None if unlimited with no leading zeros.
        :param start_num The number to begin accessioning if this is the first time.
        """
        op: AbstractAccessionServiceOperator
        if not prefix and not suffix:
            op = AccessionNextBarcode()
        else:
            op = AccessionGlobalPrefixSuffix(prefix, suffix, num_digits, start_num)
        descriptor = AccessionServiceDescriptor(op.default_accessor_name, op, num_ids, None, None)
        return self.accession_in_batch(descriptor)



import io
import warnings
import zipfile

import pandas
from numpy import dtype
from pandas import DataFrame
from sapiopylib.rest.pojo.webhook.ClientCallbackRequest import FilePromptRequest, MultiFilePromptRequest, \
    WriteFileRequest, MultiFileRequest
from sapiopylib.rest.pojo.webhook.ClientCallbackResult import FilePromptResult, MultiFilePromptResult
from sapiopylib.rest.pojo.webhook.WebhookContext import SapioWebhookContext
from sapiopylib.rest.pojo.webhook.WebhookResult import SapioWebhookResult

from sapiopycommons.general.exceptions import SapioUserErrorException


# FR-46064 - Initial port of PyWebhookUtils to sapiopycommons.
# FR-46716 - Add comments noting that some functions are deprecated in favor of CallbackUtil.
class FileUtil:
    """
    Utilities for the handling of files, including the requesting of files from the user and the parsing of files into
    tokenized lists. Makes use of Pandas DataFrames for any file parsing purposes.
    """
    @staticmethod
    def tokenize_csv(file_bytes: bytes, required_headers: list[str] | None = None, header_row_index: int | None = 0,
                     seperator: str = ",", *, encoding: str | None = None, encoding_error: str | None = "strict",
                     exception_on_empty: bool = True) -> tuple[list[dict[str, str]], list[list[str]]]:
        """
        Tokenize a CSV file. The provided file must be uniform. That is, if row 1 has 10 cells, all the rows in the file
        must have 10 cells. Otherwise, the Pandas parser throws a tokenizer exception.

        :param file_bytes: The bytes of the CSV to be parsed.
        :param required_headers: The headers that must be present in the file. If a provided header is missing, raises
            a user error exception.
        :param header_row_index: The row index in the file that the headers are located at. Everything above the header
            row is returned in the metadata list. If input is None, then no row is considered to be the header row,
            meaning that required headers are also ignored if any are provided. By default, the first row (0th index)
            is assumed to be the header row.
        :param seperator: The character that separates cells in the table.
        :param encoding: The encoding used to read the given file bytes. If not provided, uses utf-8. If your file
            contains a non-utf-8 character, then a UnicodeDecodeError will be thrown. If this happens, consider using
            ISO-8859-1 as the encoding, or investigate what encoding would handle the characters in your file.
        :param encoding_error: The error handling behavior if an encoding error is encountered. By default, the behavior
            is "strict", meaning that encoding errors raise an exception. Change this to "ignore" to skip over invalid
            characters or "replace" to replace invalid characters with a ? character. For a full list of options, see
            https://docs.python.org/3/library/codecs.html#error-handlers
        :param exception_on_empty: Throw a user error exception if the provided file bytes result in an empty list in
            the first element of the returned tuple.
        :return: The CSV parsed into a list of dicts where each dict is a row, mapping the headers to the cells for
            that row. Also returns a list of each row above the headers (the metadata), parsed into a list of each cell.
            If the header row index is 0 or None, this list will be empty.
        """
        # Parse the file bytes into two DataFrames. The first is metadata of the file located above the header row,
        # while the second is the body of the file below the header row.
        file_body, file_metadata = FileUtil.csv_to_data_frames(file_bytes, header_row_index, seperator,
                                                               encoding=encoding, encoding_error=encoding_error)
        # Parse the metadata from above the header row index into a list of lists.
        metadata: list[list[str]] = FileUtil.data_frame_to_lists(file_metadata)
        # Parse the data from the file body into a list of dicts.
        rows: list[dict[str, str]] = FileUtil.data_frame_to_dicts(file_body, required_headers, header_row_index)
        if exception_on_empty and not rows:
            raise SapioUserErrorException("The provided file contains no rows of information below the headers.")
        return rows, metadata

    @staticmethod
    def tokenize_xlsx(file_bytes: bytes, required_headers: list[str] | None = None, header_row_index: int | None = 0,
                      *, exception_on_empty: bool = True) -> tuple[list[dict[str, str]], list[list[str]]]:
        """
        Tokenize an XLSX file row by row.

        :param file_bytes: The bytes of the XLSX to be parsed.
        :param required_headers: The headers that must be present in the file. If a provided header is missing, raises
            a user error exception.
        :param header_row_index: The row index in the file that the headers are located at. Everything above the header
            row is returned in the metadata list. If input is None, then no row is considered to be the header row,
            meaning that required headers are also ignored if any are provided. By default, the first row (0th index)
            is assumed to be the header row.
        :param exception_on_empty: Throw a user error exception if the provided file bytes result in an empty list in
            the first element of the returned tuple.
        :return: The XLSX parsed into a list of dicts where each dict is a row, mapping the headers to the cells for
            that row. Also returns a list of each row above the headers (the metadata), parsed into a list of each cell.
            If the header row index is 0 or None, this list will be empty.
        """
        # Parse the file bytes into two DataFrames. The first is metadata of the file located above the header row,
        # while the second is the body of the file below the header row.
        file_body, file_metadata = FileUtil.xlsx_to_data_frames(file_bytes, header_row_index)
        # Parse the metadata from above the header row index into a list of lists.
        metadata: list[list[str]] = FileUtil.data_frame_to_lists(file_metadata)
        # Parse the data from the file body into a list of dicts.
        rows: list[dict[str, str]] = FileUtil.data_frame_to_dicts(file_body, required_headers, header_row_index)
        if exception_on_empty and not rows:
            raise SapioUserErrorException("The provided file contains no rows of information below the headers.")
        return rows, metadata

    @staticmethod
    def csv_to_data_frames(file_bytes: bytes, header_row_index: int | None = 0, seperator: str = ",",
                           *, encoding: str | None = None, encoding_error: str | None = "strict") \
            -> tuple[DataFrame, DataFrame | None]:
        """
        Parse the file bytes for a CSV into DataFrames. The provided file must be uniform. That is, if row 1 has 10
        cells, all the rows in the file must have 10 cells. Otherwise, the Pandas parser throws a tokenizer exception.

        :param file_bytes: The bytes of the CSV to be parsed.
        :param header_row_index: The row index in the file that the headers are located at. Everything above the header
            row is returned in the metadata list. If input is None, then no row is considered to be the header row,
            meaning that required headers are also ignored if any are provided. By default, the first row (0th index)
            is assumed to be the header row.
        :param seperator: The character that separates cells in the table.
        :param encoding: The encoding used to read the given file bytes. If not provided, uses utf-8. If your file
            contains a non-utf-8 character, then a UnicodeDecodeError will be thrown. If this happens, consider using
            ISO-8859-1 as the encoding, or investigate what encoding would handle the characters in your file.
        :param encoding_error: The error handling behavior if an encoding error is encountered. By default, the behavior
            is "strict", meaning that encoding errors raise an exception. Change this to "ignore" to skip over invalid
            characters or "replace" to replace invalid characters with a ? character. For a full list of options, see
            https://docs.python.org/3/library/codecs.html#error-handlers
        :return: A tuple of two DataFrames. The first is the frame for the CSV table body, while the second is for the
            metadata from above the header row, or None if there is no metadata.
        """
        file_metadata: DataFrame | None = None
        if header_row_index is not None and header_row_index > 0:
            with io.BytesIO(file_bytes) as file_io:
                # The metadata DataFrame has no headers and only consists of the rows above the header row index.
                # Therefore, we skip every row including and past the header. Don't skip blank rows, as skipping them
                # can throw off the header row index.
                file_metadata = pandas.read_csv(file_io, header=None, dtype=dtype(str),
                                                skiprows=lambda x: x >= header_row_index,
                                                skip_blank_lines=False, sep=seperator, encoding=encoding,
                                                encoding_errors=encoding_error)
        with io.BytesIO(file_bytes) as file_io:
            # The use of the dtype argument is to ensure that everything from the file gets read as a string. Added
            # because some numerical values would get ".0" appended to them, even when casting the DataFrame cell to a
            # string.
            file_body: DataFrame = pandas.read_csv(file_io, header=header_row_index, dtype=dtype(str),
                                                   skip_blank_lines=False, sep=seperator, encoding=encoding)

        return file_body, file_metadata

    @staticmethod
    def xlsx_to_data_frames(file_bytes: bytes, header_row_index: int | None = 0) \
            -> tuple[DataFrame, DataFrame | None]:
        """
        Parse the file bytes for an XLSX into DataFrames.

        :param file_bytes: The bytes of the XLSX to be parsed.
        :param header_row_index: The row index in the file that the headers are located at. Everything above the header
            row is returned in the metadata list. If input is None, then no row is considered to be the header row,
            meaning that required headers are also ignored if any are provided. By default, the first row (0th index)
            is assumed to be the header row.
        :return: A tuple of two DataFrames. The first is the frame for the XLSX table body, while the second is for the
            metadata from above the header row, or None if there is no metadata.
        """
        file_metadata: DataFrame | None = None
        if header_row_index is not None and header_row_index > 0:
            with io.BytesIO(file_bytes) as file_io:
                # The metadata DataFrame has no headers and only consists of the rows above the header row index.
                # Therefore, we skip every row including and past the header.
                file_metadata = pandas.read_excel(file_io, header=None, dtype=dtype(str),
                                                  skiprows=lambda x: x >= header_row_index)
        with io.BytesIO(file_bytes) as file_io:
            # The use of the dtype argument is to ensure that everything from the file gets read as a string. Added
            # because some numerical values would get ".0" appended to them, even when casting the DataFrame cell to a
            # string.
            file_body: DataFrame = pandas.read_excel(file_io, header=header_row_index, dtype=dtype(str))

        return file_body, file_metadata

    @staticmethod
    def data_frame_to_lists(data_frame: DataFrame) -> list[list[str]]:
        """
        Parse a Pandas DataFrame to a list of lists. Each outer list is a row, while each inner list is the cells
        for that row.

        :param data_frame: The DataFrame to be parsed.
        :return: The input DataFrame parsed into a row-wise list of lists.
        """
        if data_frame is None:
            return []
        # DataFrames are oriented column-wise instead of row-wise. Read down each column before moving to the next.
        rows: list[list[str]] = []
        for column_key in data_frame.columns:
            column_rows = list(data_frame.get(column_key))
            # If this is the first column (meaning the rows list is empty), initialize the lists for each row.
            if not rows:
                rows.extend([[] for _ in range(len(column_rows))])
            for value, row in zip(column_rows, rows):
                # Strip each value just in case there is leading or trailing whitespace.
                value = str(value).strip()
                # Empty cells get added to the list as None.
                if value == "nan":
                    value = None
                row.append(value)
        return rows

    @staticmethod
    def data_frame_to_dicts(data_frame: DataFrame,
                            required_headers: list[str] = None,
                            header_row_index: int | None = 0) -> list[dict[str, str]]:
        """
        Parse a Pandas DataFrame to a list of dicts. Each list is a row, while each dict is the cells from that row
        keyed by the column header that the cell is under. Capable of requiring that certain headers are present.

        The names of the data_frame.columns values are expected to be the header names.

        :param data_frame: The DataFrame to be parsed.
        :param required_headers: The headers that must be present in the DataFrame. If a header is missing, raises an
            exception. If no headers are provided, doesn't do any enforcement.
        :param header_row_index: The row index in the file that the headers are located at. Only used ot tell the
            user where the headers are expected to be if a required header is not found, given that the input DataFrame
            was created from a parsed file.
        :return: The input DataFrame parsed into a row-wise list of dicts.
        """
        if data_frame is None:
            return []
        # DataFrames are oriented column-wise instead of row-wise. Read down each column before moving to the next.
        rows: list[dict[str, str]] = []
        witnessed_headers: list[str] = []
        for header in data_frame.columns:
            column_rows = list(data_frame.get(header))
            witnessed_headers.append(header)
            # If this is the first column (meaning the rows list is empty), initialize the lists for each row.
            if not rows:
                rows.extend([{} for _ in range(len(column_rows))])
            for value, row in zip(column_rows, rows):
                # Strip each value just in case there is leading or trailing whitespace.
                value = str(value).strip()
                # Empty cells get added to the list as None.
                if value == "nan":
                    value = None
                row.update({header: value})

        # Warn about improper headers. Confirm that the header contains each of the header column names that we want.
        if required_headers is not None:
            # FR-46702: Report all missing headers instead of only the first missing header.
            missing_headers: list[str] = []
            for required in required_headers:
                if required not in witnessed_headers:
                    missing_headers.append("\"" + required + "\"")
            if missing_headers:
                at_row = " at row " + str(header_row_index + 1) if header_row_index is not None else ""
                raise SapioUserErrorException(f"Incorrect file headers or incorrectly formatted table. Header(s) "
                                              f"{', '.join(missing_headers)} not found{at_row}.")

        return rows

    @staticmethod
    def csv_to_xlsx(file_data: bytes | str) -> bytes:
        """
        Convert a CSV file into an XLSX file.

        :param file_data: The CSV file to be converted.
        :return: The bytes of the CSV file converted to an XLSX file.
        """
        with (io.BytesIO(file_data.encode() if isinstance(file_data, str) else file_data)) as csv:
            # Setting header to false makes pandas read the CSV as-is.
            data_frame = pandas.read_csv(csv, sep=",", header=None)

        with io.BytesIO() as output:
            with pandas.ExcelWriter(output, engine='xlsxwriter') as writer:
                # Setting header and index to false makes the CSV convert to an XLSX as-is.
                data_frame.to_excel(writer, sheet_name='Sheet1', header=False, index=False)
            xlsx_data = output.getvalue()
        return xlsx_data

    @staticmethod
    def csv_to_xls(file_data: bytes | str, delimiter: str = ",", newline: str = "\r\n") -> bytes:
        """
        Convert the bytes or string of a .csv file to .xls bytes.

        :param file_data: The .csv bytes or string to convert.
        :param delimiter: The delimiter character separating columns, with "," being the default.
        :param newline: The newline character(s) separating rows, with "\r\n" being the default.
        :return: The bytes of the new .xls file.
        """
        # Import the libraries we'll need locally since we won't be using them anywhere else in the class.
        from xlwt import Workbook, Worksheet

        # Create an Excel workbook along with a worksheet, which is where the data will be written.
        workbook: Workbook = Workbook()
        sheet: Worksheet = workbook.add_sheet("Sheet1")

        # Make sure the file data is in a string format so that we can work with it.
        formatted_data: str = bytes.decode(file_data, "utf-8") if isinstance(file_data, bytes) else file_data

        # Write each row of the file to the .xls sheet.
        rows: list[str] = formatted_data.split(newline)
        for i, row in enumerate(rows):
            values: list[str] = row.split(delimiter)
            for j, value in enumerate(values):
                sheet.write(i, j, value)

        # Save the worksheet data to the byte buffer and return the bytes.
        with io.BytesIO() as buffer:
            workbook.save(buffer)
            file_bytes: bytes = buffer.getvalue()
        return file_bytes

    @staticmethod
    def zip_files(files: dict[str, str | bytes]) -> bytes:
        """
        Create a zip file for a collection of files.

        :param files: A dictionary of file name to file data as a string or bytes.
        :return: The bytes for a zip file containing the input files.
        """
        zip_buffer: io.BytesIO = io.BytesIO()
        with zipfile.ZipFile(zip_buffer, "w", zipfile.ZIP_DEFLATED) as zip_file:
            for file_name, file_data in files.items():
                zip_file.writestr(file_name, file_data)
        return zip_buffer.getvalue()

    # Deprecated functions:

    # FR-46097 - Add write file request shorthand functions to FileUtil.
    @staticmethod
    def write_file(file_name: str, file_bytes: bytes, *, request_context: str | None = None) -> SapioWebhookResult:
        """
        DEPRECATED: Make use of CallbackUtil as of 24.5.

        Send a file to the client.

        The calling webhook must catch the WriteFileResult that the client will send back.

        :param file_name: The name of the file.
        :param file_bytes: The bytes of the file.
        :param request_context: Context that will be returned to the webhook server in the client callback result.
        :return: A SapioWebhookResult with the write request as its client callback request.
        """
        warnings.warn("FileUtil.write_file is deprecated as of 24.5+. Use CallbackUtil.write_file instead.",
                      DeprecationWarning)
        return SapioWebhookResult(True, client_callback_request=WriteFileRequest(file_bytes, file_name,
                                                                                 request_context))

    @staticmethod
    def write_files(files: dict[str, bytes], *, request_context: str | None = None) -> SapioWebhookResult:
        """
        DEPRECATED: Make use of CallbackUtil as of 24.5.

        Send a collection of files to the client.

        The calling webhook must catch the MultiFileResult that the client will send back.

        :param files: A dictionary of files names to file bytes.
        :param request_context: Context that will be returned to the webhook server in the client callback result.
        :return: A SapioWebhookResult with the write request as its client callback request.
        """
        warnings.warn("FileUtil.write_files is deprecated as of 24.5+. Use CallbackUtil.write_file instead.",
                      DeprecationWarning)
        return SapioWebhookResult(True, client_callback_request=MultiFileRequest(files, request_context))

    @staticmethod
    def request_file(context: SapioWebhookContext, title: str, exts: list[str] = None,
                     *, request_context: str | None = None) \
            -> tuple[SapioWebhookResult | None, str | None, bytes | None]:
        """
        DEPRECATED: Make use of CallbackUtil as of 24.5.

        Request a single file from the user. This function handles the entire client callback interaction for the
        requesting of the file, including if the user cancels the file upload prompt.

        The first time this method is called in the course of an interaction with the client, it will return a webhook
        result containing the client callback to request the file. The second time it is called, it will return the
        file name and bytes from the callback result.

        :param context: The current webhook context.
        :param title: The title of the file prompt dialog.
        :param exts: The allowable file extensions of the uploaded file. If blank, any file can be uploaded. Throws an
            exception if an incorrect file extension is provided.
        :param request_context: Context that will be returned to the webhook server in the client callback result.
        :return: A tuple with the following elements.
            0 - A webhook result that contains a file prompt if this is the first interaction with this request.
            May also contain a result that will terminate the client interaction if the user canceled the prompt.
            1 - The file name of the requested file if the user provided one.
            2 - The file bytes of the requested file if the user provided one.
        """
        warnings.warn("FileUtil.request_file is deprecated as of 24.5+. Use CallbackUtil.request_file instead.",
                      DeprecationWarning)
        client_callback = context.client_callback_result
        result_context: str | None = client_callback.callback_context_data if client_callback else None
        # If the user cancels, terminate the interaction.
        if client_callback is not None and client_callback.user_cancelled:
            return SapioWebhookResult(True), None, None
        # If no extensions were provided, use an empty list for the extensions instead.
        if exts is None:
            exts = []

        # If the client callback isn't a FilePromptResult, then it's either None or some other callback result, meaning
        # we need to send a new request. We may also send a new request if the client callback result is a
        # FilePromptResult, but its callback context doesn't match the provided callback context, meaning it's a
        # result from a different call to request_file.
        is_file_result = isinstance(client_callback, FilePromptResult)
        if not is_file_result or (is_file_result and result_context != request_context):
            prompt = FilePromptRequest(dialog_title=title, file_extension=",".join(exts),
                                       callback_context_data=request_context)
            return SapioWebhookResult(True, client_callback_request=prompt), None, None

        # Get the file from the result. Enforce that the provided data isn't empty, and that the file path ends in
        # one of the allowed extensions.
        # noinspection PyTypeChecker
        result: FilePromptResult = client_callback
        file_path: str | None = result.file_path
        file_bytes: bytes | None = result.file_bytes
        FileUtil.__verify_file(file_path, file_bytes, exts)
        return None, file_path, file_bytes

    @staticmethod
    def request_files(context: SapioWebhookContext, title: str, exts: list[str] = None,
                      *, request_context: str | None = None) \
            -> tuple[SapioWebhookResult | None, dict[str, bytes] | None]:
        """
        DEPRECATED: Make use of CallbackUtil as of 24.5.

        Request multiple files from the user. This function handles the entire client callback interaction for the
        requesting of the files, including if the user cancels the file upload prompt.

        The first time this method is called in the course of an interaction with the client, it will return a webhook
        result containing the client callback to request the files. The second time it is called, it will return each
        file name and bytes from the callback result.

        :param context: The current webhook context.
        :param title: The title of the file prompt dialog.
        :param exts: The allowable file extensions of the uploaded file. If blank, any file can be uploaded. Throws an
            exception if an incorrect file extension is provided.
        :param request_context: Context that will be returned to the webhook server in the client callback result.
        :return: A tuple with the following elements.
            0 - A webhook result that contains a file prompt if this is the first interaction with this request.
            May also contain a result that will terminate the client interaction if the user canceled the prompt.
            1 - A dictionary that maps the file names to the file bytes for each provided file.
        """
        warnings.warn("FileUtil.request_files is deprecated as of 24.5+. Use CallbackUtil.request_files instead.",
                      DeprecationWarning)
        client_callback = context.client_callback_result
        result_context: str | None = client_callback.callback_context_data if client_callback else None
        # If the user cancels, terminate the interaction.
        if client_callback is not None and client_callback.user_cancelled:
            return SapioWebhookResult(True), None
        # If no extensions were provided, use an empty list for the extensions instead.
        if exts is None:
            exts = []

        # If the client callback isn't a MultiFilePromptResult, then it's either None or some other callback result,
        # meaning we need to send a new request. We may also send a new request if the client callback result is a
        # MultiFilePromptResult, but its callback context doesn't match the provided callback context, meaning it's a
        # result from a different call to request_file.
        is_file_result = isinstance(client_callback, MultiFilePromptResult)
        if not is_file_result or (is_file_result and result_context != request_context):
            prompt = MultiFilePromptRequest(dialog_title=title, file_extension=",".join(exts),
                                            callback_context_data=request_context)
            return SapioWebhookResult(True, client_callback_request=prompt), None

        # Get the files from the result. Enforce that the provided data isn't empty, and that the file paths end in
        # one of the allowed extensions.
        # noinspection PyTypeChecker
        result: MultiFilePromptResult = client_callback
        for file_path, file_bytes in result.files.items():
            FileUtil.__verify_file(file_path, file_bytes, exts)
        return None, result.files

    @staticmethod
    def __verify_file(file_path: str, file_bytes: bytes, allowed_extensions: list[str]):
        """
        Verify that the provided file was read (i.e. the file path and file bytes aren't None or empty) and that it
        has the correct file extension. Raises a user error exception if something about the file is incorrect.

        :param file_path: The name of the file to verify.
        :param file_bytes: The bytes of the file to verify.
        :param allowed_extensions: The file extensions that the file path is allowed to have.
        """
        if file_path is None or len(file_path) == 0 or file_bytes is None or len(file_bytes) == 0:
            raise SapioUserErrorException("Empty file provided or file unable to be read.")
        if len(allowed_extensions) != 0:
            matches: bool = False
            for ext in allowed_extensions:
                if file_path.endswith("." + ext.lstrip(".")):
                    matches = True
                    break
            if matches is False:
                raise SapioUserErrorException("Unsupported file type. Expecting the following extension(s): "
                                              + (",".join(allowed_extensions)))



import warnings

from sapiopylib.rest.DataMgmtService import DataMgmtServer
from sapiopylib.rest.pojo.datatype.DataType import DataTypeDefinition
from sapiopylib.rest.pojo.datatype.FieldDefinition import VeloxStringFieldDefinition, AbstractVeloxFieldDefinition, \
    VeloxIntegerFieldDefinition, VeloxDoubleFieldDefinition
from sapiopylib.rest.pojo.webhook.ClientCallbackRequest import FormEntryDialogRequest, OptionDialogRequest, \
    TableEntryDialogRequest, ListDialogRequest, DataRecordSelectionRequest
from sapiopylib.rest.pojo.webhook.WebhookContext import SapioWebhookContext
from sapiopylib.rest.pojo.webhook.WebhookResult import SapioWebhookResult
from sapiopylib.rest.utils.FormBuilder import FormBuilder

from sapiopycommons.general.aliases import SapioRecord, AliasUtil, FieldMap
from sapiopycommons.general.exceptions import SapioException


# FR-46064 - Initial port of PyWebhookUtils to sapiopycommons.
# FR-46097 - Greatly expand the options that PopupUtil provides. (Originally just had two OptionDialogRequest
# and one FormEntryDialogRequest methods.)
# CR-46332 - For any functions that use temporary data types, set the data type name, display name, and plural display
# name in the form builder.
# FR-46716 - Add comments noting that this class is deprecated in favor of CallbackUtil.
class PopupUtil:
    """
    DEPRECATED: Make use of CallbackUtil as of 24.5.

    Methods for creating boilerplate SapioWebhookResults with client callback requests to create popup dialogs.
    """
    @staticmethod
    def form_popup(title: str, msg: str, fields: list[AbstractVeloxFieldDefinition], values: FieldMap = None,
                   column_positions: dict[str, tuple[int, int]] = None, data_type: str = "Default",
                   *, display_name: str | None = None, plural_display_name: str | None = None,
                   request_context: str | None = None) -> SapioWebhookResult:
        """
        DEPRECATED: Make use of CallbackUtil as of 24.5.

        Create a basic form entry dialog.

        The calling webhook must catch the FormEntryDialogResult that the client will send back.

        :param title: The title of the dialog.
        :param msg: The message to display at the top of the form.
        :param fields: The definitions of the fields to display in the form. Fields will be displayed in the order they
            are provided in this list.
        :param values: Sets the default values of the fields.
        :param column_positions: If a tuple is provided for a field name, alters that field's column position and column
            span. (Field order is still determined by the fields list.)
        :param data_type: The data type name for the temporary data type that will be created for this form.
        :param display_name: The display name for the temporary data type. If not provided, defaults to the data type
            name.
        :param plural_display_name: The plural display name for the temporary data type. If not provided, defaults to
            the display name + "s".
        :param request_context: Context that will be returned to the webhook server in the client callback result.
        :return: A SapioWebhookResult with the popup as its client callback request.
        """
        warnings.warn("PopupUtil is deprecated as of 24.5+. Use CallbackUtil instead.", DeprecationWarning)
        if display_name is None:
            display_name = data_type
        if plural_display_name is None:
            plural_display_name = display_name + "s"
        builder = FormBuilder(data_type, display_name, plural_display_name)
        for field_def in fields:
            field_name = field_def.data_field_name
            if values and hasattr(field_def, "default_value"):
                field_def.default_value = values.get(field_name)
            column: int = 0
            span: int = 4
            if column_positions and field_name in column_positions:
                position = column_positions.get(field_name)
                column = position[0]
                span = position[1]
            builder.add_field(field_def, column, span)
        callback = FormEntryDialogRequest(title, msg, builder.get_temporary_data_type(),
                                          callback_context_data=request_context)
        return SapioWebhookResult(True, client_callback_request=callback)

    @staticmethod
    def record_form_popup(context: SapioWebhookContext,
                          title: str, msg: str, fields: list[str], record: SapioRecord,
                          column_positions: dict[str, tuple[int, int]] = None, editable: bool | None = True,
                          *, request_context: str | None = None) -> SapioWebhookResult:
        """
        DEPRECATED: Make use of CallbackUtil as of 24.5.

        Create a basic form dialog for a record with displayed fields from the record data type.

        Makes webservice calls to get the data field and type definitions of the given data type.
        The calling webhook must catch the FormEntryDialogResult that the client will send back.

        :param context: The current webhook context.
        :param title: The title of the dialog.
        :param msg: The message to display in the dialog.
        :param fields: The data field names of the fields from the record to display in the form. Fields will be
            displayed in the order they are provided in this list.
        :param record: The record to display the values of.
        :param column_positions: If a tuple is provided for a field name, alters that field's column position and column
            span. (Field order is still determined by the fields list.)
        :param editable: If true, all fields are displayed as editable. If false, all fields are displayed as
            uneditable. If none, only those fields that are defined as editable by the data designer will be editable.
        :param request_context: Context that will be returned to the webhook server in the client callback result.
        :return: A SapioWebhookResult with the popup as its client callback request.
        """
        warnings.warn("PopupUtil is deprecated as of 24.5+. Use CallbackUtil instead.", DeprecationWarning)
        # Get the field definitions of the data type.
        data_type: str = record.data_type_name
        type_man = DataMgmtServer.get_data_type_manager(context.user)
        type_def: DataTypeDefinition = type_man.get_data_type_definition(data_type)
        field_defs: dict[str, AbstractVeloxFieldDefinition] = {x.data_field_name: x for x in
                                                               type_man.get_field_definition_list(data_type)}

        # Build the form using only those fields that are desired.
        builder = FormBuilder(data_type, type_def.display_name, type_def.plural_display_name)
        for field_name in fields:
            field_def = field_defs.get(field_name)
            if field_def is None:
                raise SapioException(f"No field of name \"{field_name}\" in field definitions of type \"{data_type}\"")
            if editable is not None:
                field_def.editable = editable
            field_def.visible = True
            if hasattr(field_def, "default_value"):
                field_def.default_value = record.get_field_value(field_name)
            column: int = 0
            span: int = 4
            if column_positions and field_name in column_positions:
                position = column_positions.get(field_name)
                column = position[0]
                span = position[1]
            builder.add_field(field_def, column, span)
        temp_type_def = builder.get_temporary_data_type()
        callback = FormEntryDialogRequest(title, msg, temp_type_def,
                                          callback_context_data=request_context)
        return SapioWebhookResult(True, client_callback_request=callback)

    @staticmethod
    def string_field_popup(title: str, msg: str, field_name: str, default_value: str | None = None,
                           max_length: int | None = None, editable: bool = True, data_type: str = "Default",
                           *, display_name: str | None = None, plural_display_name: str | None = None,
                           request_context: str | None = None, **kwargs) -> SapioWebhookResult:
        """
        DEPRECATED: Make use of CallbackUtil as of 24.5.

        Create a form dialog with a single string field. May take additional parameters to be passed to the string field
        definition.

        The calling webhook must catch the FormEntryDialogResult that the client will send back.

        :param title: The title of the dialog.
        :param msg: The message to display in the dialog.
        :param field_name: The name and display name of the string field.
        :param default_value: The default value to place into the string field, if any.
        :param max_length: The max length of the string value. If not provided, uses the length of the default value.
            If neither this or a default value are not provided, defaults to 100 characters.
        :param editable: Whether the user may edit the contents of the string field.
        :param data_type: The data type name for the temporary data type that will be created for this form.
        :param display_name: The display name for the temporary data type. If not provided, defaults to the data type
            name.
        :param plural_display_name: The plural display name for the temporary data type. If not provided, defaults to
            the display name + "s".
        :param request_context: Context that will be returned to the webhook server in the client callback result.
        :return: A SapioWebhookResult with the popup as its client callback request.
        """
        warnings.warn("PopupUtil is deprecated as of 24.5+. Use CallbackUtil instead.", DeprecationWarning)
        if max_length is None:
            max_length = len(default_value) if default_value else 100
        string_field = VeloxStringFieldDefinition(data_type, field_name, field_name, default_value=default_value,
                                                  max_length=max_length, editable=editable, **kwargs)
        return PopupUtil.form_popup(title, msg, [string_field], data_type=data_type, display_name=display_name,
                                    plural_display_name=plural_display_name, request_context=request_context)

    @staticmethod
    def integer_field_popup(title: str, msg: str, field_name: str, default_value: int = None, min_value: int = -10000,
                            max_value: int = 10000, data_type: str = "Default", editable: bool = True,
                            *, display_name: str | None = None, plural_display_name: str | None = None,
                            request_context: str | None = None, **kwargs) -> SapioWebhookResult:
        """
        DEPRECATED: Make use of CallbackUtil as of 24.5.

        Create a form dialog with a single integer field. May take additional parameters to be passed to the integer
        field definition.

        The calling webhook must catch the FormEntryDialogResult that the client will send back.

        :param title: The title of the dialog.
        :param msg: The message to display in the dialog.
        :param field_name: The name and display name of the integer field.
        :param default_value: The default value to place into the integer field. If not provided, defaults to the 0 or
            the minimum value, whichever is higher.
        :param min_value: The minimum allowed value of the input.
        :param max_value: The maximum allowed value of the input.
        :param data_type: The data type name for the temporary data type that will be created for this form.
        :param editable: Whether the user may edit the contents of the integer field.
        :param display_name: The display name for the temporary data type. If not provided, defaults to the data type
            name.
        :param plural_display_name: The plural display name for the temporary data type. If not provided, defaults to
            the display name + "s".
        :param request_context: Context that will be returned to the webhook server in the client callback result.
        :return: A SapioWebhookResult with the popup as its client callback request.
        """
        warnings.warn("PopupUtil is deprecated as of 24.5+. Use CallbackUtil instead.", DeprecationWarning)
        if default_value is None:
            default_value = max(0, min_value)
        integer_field = VeloxIntegerFieldDefinition(data_type, field_name, field_name, default_value=default_value,
                                                    min_value=min_value, max_value=max_value, editable=editable,
                                                    **kwargs)
        return PopupUtil.form_popup(title, msg, [integer_field], data_type=data_type, display_name=display_name,
                                    plural_display_name=plural_display_name, request_context=request_context)

    @staticmethod
    def double_field_popup(title: str, msg: str, field_name: str, default_value: float = None,
                           min_value: float = -10000000, max_value: float = 100000000,
                           data_type: str = "Default", editable: bool = True, *, display_name: str | None = None,
                           plural_display_name: str | None = None, request_context: str | None = None, **kwargs) \
            -> SapioWebhookResult:
        """
        DEPRECATED: Make use of CallbackUtil as of 24.5.

        Create a form dialog with a single double field. May take additional parameters to be passed to the double
        field definition.

        The calling webhook must catch the FormEntryDialogResult that the client will send back.

        :param title: The title of the dialog.
        :param msg: The message to display in the dialog.
        :param field_name: The name and display name of the double field.
        :param default_value: The default value to place into the double field. If not provided, defaults to the 0 or
            the minimum value, whichever is higher.
        :param min_value: The minimum allowed value of the input.
        :param max_value: The maximum allowed value of the input.
        :param data_type: The data type name for the temporary data type that will be created for this form.
        :param editable: Whether the user may edit the contents of the double field.
        :param display_name: The display name for the temporary data type. If not provided, defaults to the data type
            name.
        :param plural_display_name: The plural display name for the temporary data type. If not provided, defaults to
            the display name + "s".
        :param request_context: Context that will be returned to the webhook server in the client callback result.
        :return: A SapioWebhookResult with the popup as its client callback request.
        """
        warnings.warn("PopupUtil is deprecated as of 24.5+. Use CallbackUtil instead.", DeprecationWarning)
        if default_value is None:
            default_value = min_value
        double_field = VeloxDoubleFieldDefinition(data_type, field_name, field_name, default_value=default_value,
                                                  min_value=min_value, max_value=max_value, editable=editable, **kwargs)
        return PopupUtil.form_popup(title, msg, [double_field], data_type=data_type, display_name=display_name,
                                    plural_display_name=plural_display_name, request_context=request_context)

    @staticmethod
    def table_popup(title: str, msg: str, fields: list[AbstractVeloxFieldDefinition], values: list[FieldMap],
                    *, data_type: str = "Default", display_name: str | None = None,
                    plural_display_name: str | None = None, request_context: str | None = None) -> SapioWebhookResult:
        """
        DEPRECATED: Make use of CallbackUtil as of 24.5.

        Create a basic table entry dialog.

        The calling webhook must catch the TableEntryDialogResult that the client will send back.

        :param title: The title of the dialog.
        :param msg: The message to display at the top of the form.
        :param fields: The definitions of the fields to display as table columns. Fields will be displayed in the order
            they are provided in this list.
        :param values: The values to set for each row of the table.
        :param data_type: The data type name for the temporary data type that will be created for this table.
        :param display_name: The display name for the temporary data type. If not provided, defaults to the data type
            name.
        :param plural_display_name: The plural display name for the temporary data type. If not provided, defaults to
            the display name + "s".
        :param request_context: Context that will be returned to the webhook server in the client callback result.
        :return: A SapioWebhookResult with the popup as its client callback request.
        """
        warnings.warn("PopupUtil is deprecated as of 24.5+. Use CallbackUtil instead.", DeprecationWarning)
        if display_name is None:
            display_name = data_type
        if plural_display_name is None:
            plural_display_name = display_name + "s"
        builder = FormBuilder(data_type, display_name, plural_display_name)
        for column in fields:
            builder.add_field(column)
        temp_type_def = builder.get_temporary_data_type()
        callback = TableEntryDialogRequest(title, msg, temp_type_def, values,
                                           callback_context_data=request_context)
        return SapioWebhookResult(True, client_callback_request=callback)

    @staticmethod
    def record_table_popup(context: SapioWebhookContext,
                           title: str, msg: str, fields: list[str], records: list[SapioRecord],
                           editable: bool | None = True, *, request_context: str | None = None) -> SapioWebhookResult:
        """
        DEPRECATED: Make use of CallbackUtil as of 24.5.

        Create a table dialog for a list of record where the columns are specific fields from the record data type.

        Makes webservice calls to get the data field and type definitions of the given data type.
        The calling webhook must catch the TableEntryDialogResult that the client will send back.

        :param context: The current webhook context
        :param title: The title of the dialog.
        :param msg: The message to display in the dialog.
        :param records: The records to display as rows in the table.
        :param fields: The names of the fields to display as columns in the table. Fields will be displayed in the order
            they are provided in this list.
        :param editable: If true, all fields are displayed as editable. If false, all fields are displayed as
            uneditable. If none, only those fields that are defined as editable by the data designer will be editable.
        :param request_context: Context that will be returned to the webhook server in the client callback result.
        :return: A SapioWebhookResult with the popup as its client callback request.
        """
        warnings.warn("PopupUtil is deprecated as of 24.5+. Use CallbackUtil instead.", DeprecationWarning)
        if not records:
            raise SapioException("No records provided.")
        data_types: set[str] = {x.data_type_name for x in records}
        if len(data_types) > 1:
            raise SapioException("Multiple data type names encountered in records list for record table popup.")
        data_type: str = data_types.pop()
        # Get the field maps from the records.
        field_map_list: list[FieldMap] = AliasUtil.to_field_map_lists(records)
        # Get the field definitions of the data type.
        type_man = DataMgmtServer.get_data_type_manager(context.user)
        type_def: DataTypeDefinition = type_man.get_data_type_definition(data_type)
        field_defs: dict[str, AbstractVeloxFieldDefinition] = {x.data_field_name: x for x in
                                                               type_man.get_field_definition_list(data_type)}

        # Build the form using only those fields that are desired.
        builder = FormBuilder(data_type, type_def.display_name, type_def.plural_display_name)
        for field_name in fields:
            field_def = field_defs.get(field_name)
            if field_def is None:
                raise SapioException(f"No field of name \"{field_name}\" in field definitions of type \"{data_type}\"")
            if editable is not None:
                field_def.editable = editable
            field_def.visible = True
            # Key fields display their columns in order before all non-key fields.
            # Unmark key fields so that the column order is respected exactly as the caller provides it.
            field_def.key_field = False
            builder.add_field(field_def)
        temp_type_def = builder.get_temporary_data_type()
        callback = TableEntryDialogRequest(title, msg, temp_type_def, field_map_list,
                                           callback_context_data=request_context)
        return SapioWebhookResult(True, client_callback_request=callback)

    @staticmethod
    def record_selection_popup(context: SapioWebhookContext,
                               msg: str, fields: list[str], records: list[SapioRecord], multi_select: bool = True,
                               *, request_context: str | None = None) -> SapioWebhookResult:
        """
        DEPRECATED: Make use of CallbackUtil as of 24.5.

        Create a record selection dialog for a list of record where the columns are specific fields from the record data
        type.

        Makes webservice calls to get the data field and type definitions of the given data type.
        The calling webhook must catch the DataRecordSelectionResult that the client will send back.

        :param context: The current webhook context
        :param msg: The message to display in the dialog.
        :param records: The records to display as rows in the table.
        :param fields: The names of the fields to display as columns in the table. Fields will be displayed in the order
            they are provided in this list.
        :param multi_select: Whether the user is able to select multiple records from the list.
        :param request_context: Context that will be returned to the webhook server in the client callback result.
        :return: A SapioWebhookResult with the popup as its client callback request.
        """
        warnings.warn("PopupUtil is deprecated as of 24.5+. Use CallbackUtil instead.", DeprecationWarning)
        if not records:
            raise SapioException("No records provided.")
        data_types: set[str] = {x.data_type_name for x in records}
        if len(data_types) > 1:
            raise SapioException("Multiple data type names encountered in records list for record table popup.")
        data_type: str = data_types.pop()
        # Get the field maps from the records.
        field_map_list: list[FieldMap] = AliasUtil.to_field_map_lists(records)
        # Get the field definitions of the data type.
        type_man = DataMgmtServer.get_data_type_manager(context.user)
        type_def: DataTypeDefinition = type_man.get_data_type_definition(data_type)
        field_defs: dict[str, AbstractVeloxFieldDefinition] = {x.data_field_name: x for x in
                                                               type_man.get_field_definition_list(data_type)}

        # Build the form using only those fields that are desired.
        field_def_list: list = []
        for field_name in fields:
            field_def = field_defs.get(field_name)
            if field_def is None:
                raise SapioException(f"No field of name \"{field_name}\" in field definitions of type \"{data_type}\"")
            field_def.visible = True
            # Key fields display their columns in order before all non-key fields.
            # Unmark key fields so that the column order is respected exactly as the caller provides it.
            field_def.key_field = False
            field_def_list.append(field_def)
        callback = DataRecordSelectionRequest(type_def.display_name, type_def.plural_display_name,
                                              field_def_list, field_map_list, msg, multi_select,
                                              callback_context_data=request_context)
        return SapioWebhookResult(True, client_callback_request=callback)

    @staticmethod
    def list_popup(title: str, options: list[str], multi_select: bool = False,
                   *, request_context: str | None = None) -> SapioWebhookResult:
        """
        DEPRECATED: Make use of CallbackUtil as of 24.5.

        Create a list dialog with the given options for the user to choose from.

        The calling webhook must catch the ListDialogResult that the client will send back.

        :param title: The title of the dialog.
        :param options: The list options that the user has to choose from.
        :param multi_select: Whether the user is able to select multiple options from the list.
        :param request_context: Context that will be returned to the webhook server in the client callback result.
        :return: A SapioWebhookResult with the popup as its client callback request.
        """
        warnings.warn("PopupUtil is deprecated as of 24.5+. Use CallbackUtil instead.", DeprecationWarning)
        callback = ListDialogRequest(title, multi_select, options,
                                     callback_context_data=request_context)
        return SapioWebhookResult(True, client_callback_request=callback)

    @staticmethod
    def option_popup(title: str, msg: str, options: list[str], default_option: int = 0, user_can_cancel: bool = False,
                     *, request_context: str | None = None) -> SapioWebhookResult:
        """
        DEPRECATED: Make use of CallbackUtil as of 24.5.

        Create an option dialog with the given options for the user to choose from.

        The calling webhook must catch the OptionDialogResult that the client will send back.

        :param title: The title of the dialog.
        :param msg: The message to display in the dialog.
        :param options: The button options that the user has to choose from.
        :param default_option: The index of the option in the options list that defaults as the first choice.
        :param user_can_cancel: True if the user is able to click the X to close the dialog, returning
            user_cancelled = True in the client callback result. False if the user cannot close the dialog without
            selecting an option.
        :param request_context: Context that will be returned to the webhook server in the client callback result.
        :return: A SapioWebhookResult with the popup as its client callback request.
        """
        warnings.warn("PopupUtil is deprecated as of 24.5+. Use CallbackUtil instead.", DeprecationWarning)
        callback = OptionDialogRequest(title, msg, options, default_option, user_can_cancel,
                                       callback_context_data=request_context)
        return SapioWebhookResult(True, client_callback_request=callback)

    @staticmethod
    def ok_popup(title: str, msg: str, user_can_cancel: bool = False, *, request_context: str | None = None) \
            -> SapioWebhookResult:
        """
        DEPRECATED: Make use of CallbackUtil as of 24.5.

        Create an option dialog where the only option is "OK".

        The calling webhook must catch the OptionDialogResult that the client will send back.

        :param title: The title of the dialog.
        :param msg: The message to display in the dialog.
        :param user_can_cancel: True if the user is able to click the X to close the dialog, returning
            user_cancelled = True in the client callback result. False if the user cannot close the dialog without
            selecting an option.
        :param request_context: Context that will be returned to the webhook server in the client callback result.
        :return: A SapioWebhookResult with the popup as its client callback request.
        """
        warnings.warn("PopupUtil is deprecated as of 24.5+. Use CallbackUtil instead.", DeprecationWarning)
        return PopupUtil.option_popup(title, msg, ["OK"], 0, user_can_cancel, request_context=request_context)

    @staticmethod
    def yes_no_popup(title: str, msg: str, default_yes: bool = True, user_can_cancel: bool = False,
                     *, request_context: str | None = None) -> SapioWebhookResult:
        """
        DEPRECATED: Make use of CallbackUtil as of 24.5.

        Create an option dialog where the only options are "Yes" and "No".

        The calling webhook must catch the OptionDialogResult that the client will send back.

        :param title: The title of the dialog.
        :param msg: The message to display in the dialog.
        :param default_yes: If true, "Yes" is the default choice. Otherwise, the default choice is "No".
        :param user_can_cancel: True if the user is able to click the X to close the dialog, returning
            user_cancelled = True in the client callback result. False if the user cannot close the dialog without
            selecting an option.
        :param request_context: Context that will be returned to the webhook server in the client callback result.
        :return: A SapioWebhookResult with the popup as its client callback request.
        """
        warnings.warn("PopupUtil is deprecated as of 24.5+. Use CallbackUtil instead.", DeprecationWarning)
        return PopupUtil.option_popup(title, msg, ["Yes", "No"], 0 if default_yes else 1, user_can_cancel,
                                      request_context=request_context)

    # FR-46097 - Deprecating the three original functions for ones with briefer names. Functionality is unchanged.
    @staticmethod
    def display_form_popup(title: str, field_name: str, msg: str, data_type: str = "Popup",
                           request_context: str | None = None) -> SapioWebhookResult:
        """
        DEPRECATED: Make use of CallbackUtil as of 24.5.

        Deprecated for PopupUtil.text_field_popup.
        """
        warnings.warn("PopupUtil is deprecated as of 24.5+. Use CallbackUtil instead.", DeprecationWarning)
        return PopupUtil.string_field_popup(title, "", field_name, msg, len(msg), False, data_type,
                                            request_context=request_context, auto_size=True)

    @staticmethod
    def display_option_popup(title: str, msg: str, options: list[str], user_can_cancel: bool = False,
                             request_context: str | None = None) -> SapioWebhookResult:
        """
        DEPRECATED: Make use of CallbackUtil as of 24.5.

        Deprecated for PopupUtil.option_popup.
        """
        warnings.warn("PopupUtil is deprecated as of 24.5+. Use CallbackUtil instead.", DeprecationWarning)
        return PopupUtil.option_popup(title, msg, options, 0, user_can_cancel, request_context=request_context)

    @staticmethod
    def display_ok_popup(title: str, msg: str, request_context: str | None = None) -> SapioWebhookResult:
        """
        DEPRECATED: Make use of CallbackUtil as of 24.5.

        Deprecated for PopupUtil.ok_popup.
        """
        warnings.warn("PopupUtil is deprecated as of 24.5+. Use CallbackUtil instead.", DeprecationWarning)
        return PopupUtil.ok_popup(title, msg, False, request_context=request_context)



# Multimodal registration client
from __future__ import annotations

import io
from weakref import WeakValueDictionary

from databind.json import dumps, loads
from sapiopylib.rest.User import SapioUser

from sapiopycommons.general.exceptions import SapioException
from sapiopycommons.multimodal.multimodal_data import *


class MultiModalManager:
    _user: SapioUser

    __instances: WeakValueDictionary[SapioUser, MultiModalManager] = WeakValueDictionary()
    __initialized: bool

    def __new__(cls, user: SapioUser):
        """
        Observes singleton pattern per record model manager object.

        :param user: The user that will make the webservice request to the application.
        """
        obj = cls.__instances.get(user)
        if not obj:
            obj = object.__new__(cls)
            obj.__initialized = False
            cls.__instances[user] = obj
        return obj

    def __init__(self, user:SapioUser):
        if self.__initialized:
            return
        self._user = user
        self.__initialized = True

    def load_image_data(self, request: ImageDataRequestPojo) -> list[str]:
        """
        Loading of image data of a compound or a reaction in Sapio's unified drawing format.
        :param request:
        :return:
        """
        payload = dumps(request, ImageDataRequestPojo)
        response = self._user.plugin_post("chemistry/request_image_data",
                                          payload=payload, is_payload_plain_text=True)
        self._user.raise_for_status(response)
        return response.json()

    def load_compounds(self, request: CompoundLoadRequestPojo):
        """
        Load compounds from the provided data here.
        The compounds will not be registered but returned to you "the script".
        To complete registration, you need to call register_compounds method after obtaining result.
        """
        payload = dumps(request, CompoundLoadRequestPojo)
        response = self._user.plugin_post("chemistry/load",
                                          payload=payload, is_payload_plain_text=True)
        self._user.raise_for_status(response)
        return loads(response.text, PyMoleculeLoaderResult)

    def register_compounds(self, request: ChemRegisterRequestPojo) -> ChemCompleteImportPojo:
        """
        Register the filled compounds that are previously loaded via load_compounds operation.
        """
        payload = dumps(request, ChemRegisterRequestPojo)
        response = self._user.plugin_post("chemistry/register",
                                          payload=payload, is_payload_plain_text=True)
        self._user.raise_for_status(response)
        return loads(response.text, ChemCompleteImportPojo)

    def load_reactions(self, reaction_str: str) -> PyIndigoReactionPojo:
        """
        Load a reaction and return the loaded reaction result.
        :param reaction_str: A reaction string, in format of mrv, rxn, or smiles.
        """
        response = self._user.plugin_post("chemistry/reaction/load",
                                          payload=reaction_str, is_payload_plain_text=True)
        self._user.raise_for_status(response)
        return loads(response.text, PyIndigoReactionPojo)

    def register_reactions(self, reaction_str: str) -> DataRecord:
        """
        Register a single reaction provided.
        Note: if the rxn has already specified a 2D coordinate, it may not be recomputed when generating record image.
        :param reaction_str: The rxn of a reaction.
        :return: The registered data record. This can be a record that already exists or new.
        """
        response = self._user.plugin_post("chemistry/reaction/register",
                                          payload=reaction_str, is_payload_plain_text=True)
        self._user.raise_for_status(response)
        return loads(response.text, DataRecord)

    def search_structures(self, request: ChemSearchRequestPojo) -> ChemSearchResponsePojo:
        """
        Perform structure search against the Sapio registries.
        An error can be thrown as exception if search is structurally invalid.
        :param request: The request object containing the detailed context of this search.
        :return: The response object of the result.
        """
        payload = dumps(request, ChemSearchRequestPojo)
        response = self._user.plugin_post("chemistry/search",
                                          payload=payload, is_payload_plain_text=True)
        self._user.raise_for_status(response)
        return loads(response.text, ChemSearchResponsePojo)

    def run_multi_sequence_alignment(self, request: MultiSequenceAlignmentRequestPojo) -> list[MultiSequenceAlignmentSeqPojo]:
        """
        Run a multi-sequence alignment using the specified tool and strategy.
        :param request: The request object containing the sequences and alignment parameters. The parameters inside it can be the pojo dict of one of the options.
        :return: The result of the multi-sequence alignment.
        """
        payload = dumps(request, MultiSequenceAlignmentRequestPojo)
        response = self._user.plugin_post("bio/multisequencealignment",
                                          payload=payload, is_payload_plain_text=True)
        self._user.raise_for_status(response)
        return loads(response.text, list[MultiSequenceAlignmentSeqPojo])

    def register_bio(self, request: BioFileRegistrationRequest) -> BioFileRegistrationResponse:
        """
        Register to bioregistry of a file.
        """
        payload = dumps(request, BioFileRegistrationRequest)
        response = self._user.plugin_post("bio/register/file", payload=payload, is_payload_plain_text=True)
        self._user.raise_for_status(response)
        return loads(response.text, BioFileRegistrationResponse)

    def export_to_sdf(self, request: ChemExportSDFRequest) -> str:
        """
        Export the SDF files
        :param request: The request for exporting SDF file.
        :return: the SDF plain text data.
        """
        payload = dumps(request, ChemExportSDFRequest)
        response = self._user.plugin_post("chemistry/export_sdf", payload=payload, is_payload_plain_text=True)
        self._user.raise_for_status(response)
        gzip_base64: str = response.text
        if not gzip_base64:
            raise SapioException("Returning data from server is blank for export SDF.")
        decoded_bytes = base64.b64decode(gzip_base64)
        with io.BytesIO(decoded_bytes) as bytes_io:
            import gzip
            with gzip.GzipFile(fileobj=bytes_io, mode='rb') as f:
                ret: str = f.read().decode()
                return ret



from __future__ import annotations

import warnings
from abc import abstractmethod
from enum import Enum
from typing import Any

from sapiopycommons.general.aliases import SapioRecord, AliasUtil
from sapiopycommons.general.exceptions import SapioException
from sapiopycommons.general.time_util import TimeUtil


class FileWriter:
    """
    This class helps with the creation of character separated value files (e.g. CSVs, TSVs). You can make use of
    FileUtil.csv_to_xlsx to convert these files to xlsx files.
    """
    headers: list[str]
    body: list[list[Any]]
    delimiter: str
    line_break: str
    column_definitions: dict[str, ColumnDef]

    def __init__(self, headers: list[str], delimiter: str = ",", line_break: str = "\r\n"):
        """
        :param headers: The headers to display at the top of the file in the order in which they should appear.
        :param delimiter: The delimiter character(s) to ues between cells in the file.
        :param line_break: The character(s) to use as a line break at the end of rows.
        """
        self.headers = headers
        self.delimiter = delimiter
        self.line_break = line_break
        self.body = []
        self.column_definitions = {}

    def add_row_list(self, row: list[Any]) -> None:
        """
        Add a row of values to the file from a list. The length of the given list should be equal to the number of
        headers defined for this FileWriter.

        To be used when you just want to put together a simple file and don't want to deal with ColumnDefinitions and
        RowBundles.

        :param row: A row of values to add to the end of the file.
        """
        row_count: int = len(row)
        header_count: int = len(self.headers)
        if row_count != header_count:
            raise SapioException(f"The given list has {row_count} elements but this FileWriter has {header_count} "
                                 f"headers. The number of row elements must equal the number of headers.")
        self.body.append(row)

    def add_row_dict(self, row: dict[str, Any]) -> None:
        """
        Add a row of values to the file from a dict. The dict is expected to contain keys that match the headers of
        the file. For any header that exists for the file that doesn't have a matching key in the given dict, an empty
        string is printed.

        To be used when you just want to put together a simple file and don't want to deal with ColumnDefinitions and
        RowBundles.

        :param row: A row of values to add to the end of the file.
        """
        new_row: list[Any] = []
        for header in self.headers:
            new_row.append(row.get(header, ""))
        self.body.append(new_row)

    def add_column_definition(self, header: str, column_def: ColumnDef) -> None:
        """
        Add a new column definition to this FileWriter for a specific header.

        ColumnDefs are only used if the build_file function is provided with a list of RowBundles. Every header must
        have a column definition if this is the case.

        Custom column definitions can be created by defining a class that extends ColumnDef and implements the print
        method.

        :param column_def: A column definitions to be used to construct the file when build_file is
            called.
        :param header: The header that this column definition is for. If a header is provided that isn't in the headers
            list, the header is appended to the end of the list.
        """
        if header not in self.headers:
            self.headers.append(header)
        self.column_definitions[header] = column_def

    def add_column_definitions(self, column_defs: dict[str, ColumnDef]) -> None:
        """
        Add new column definitions to this FileWriter.

        ColumnDefs are only used if the build_file function is provided with a list of RowBundles. Every header must
        have a column definition if this is the case.

        Custom column definitions can be created by defining a class that extends ColumnDef and implements the print
        method.

        :param column_defs: A dictionary of header names to column definitions to be used to construct the file when
            build_file is called.
        """
        # For backwards compatibility purposes, if column definitions are provided as a list,
        # add them in order of appearance of the headers. This will only work if the headers are defined first, though.
        if isinstance(column_defs, list):
            warnings.warn("Adding column definitions is no longer expected as a list. Continuing to provide a list to "
                          "this function may result in undesirable behavior.", UserWarning)
            if not self.headers:
                raise SapioException("No headers provided to FileWriter before the column definitions were added.")
            for header, column_def in zip(self.headers, column_defs):
                self.column_definitions[header] = column_def
        for header, column_def in column_defs.items():
            self.add_column_definition(header, column_def)

    def build_file(self, rows: list[RowBundle] | None = None, sorter=None, reverse: bool = False) -> str:
        """
        Build the file according the information that has been given to this FileWriter. If any add_row calls were
        made, those will be listed first before any rows constructed using any given RowBundles (assuming you aren't
        sorting the rows). RowBundles can only be used if this FileWriter was also provided with ColumnDefs for mapping
        the bundles to the file.
        
        If ever a None value is countered, instead prints an empty string.

        :param rows: A list of information used to populate rows in the file. If this parameter is provided, then this
            FileWriter must have also been given column definitions to map the row information to the file with.
        :param sorter: Some function to sort the rows by before they are printed.
            See https://docs.python.org/3.10/howto/sorting.html#sortinghowto for details on the type of functions that
            can be provided.
        :param reverse: Whether the above sorter should be run in reverse.
        :return: A string of the created file. Call string.encode() to turn this into a byte array for client callbacks.
        """
        # If any column definitions have been provided, the number of column definitions and headers must be equal.
        if self.column_definitions:
            for header in self.headers:
                if header not in self.column_definitions:
                    raise SapioException(f"FileWriter has no column definition for the header {header}. If any column "
                                         f"definitions are provided, then all headers must have a column definition.")
        # If any RowBundles have been provided, there must be column definitions for mapping them to the file.
        elif rows:
            raise SapioException(f"FileWriter was given RowBundles but contains no column definitions for mapping "
                                 f"them to file contents. Either add ColumnDefs to map the RowBundles with, or use the "
                                 f"simple add_row functions.")

        file: str = self.delimiter.join(self.headers) + self.line_break
        self.__build_rows(rows)
        if sorter is not None:
            sorted(self.body, key=sorter, reverse=reverse)
        for row in self.body:
            file += self.delimiter.join([self.__str(x) for x in row]) + self.line_break
        return file

    def __build_rows(self, rows: list[RowBundle] | None = None) -> None:
        """
        Populate the FileWriter's body using the RowBundles and ColumnDefs.

        :param rows: A list of RowBundles to populate the file body with.
        """
        if not rows:
            return
        rows.sort(key=lambda x: x.index)
        for row in rows:
            new_row: list[Any] = []
            for header in self.headers:
                column = self.column_definitions[header]
                if column.may_skip and row.may_skip:
                    new_row.append("")
                else:
                    new_row.append(column.print(row))
            self.body.append(new_row)

    def __str(self, value: Any) -> str:
        """
        :param value: Some value to convert to a string.
        :return: The input value as a string. If the string of the input value contains the delimiter character, then
            the returned value is surrounded by quotation marks.
        """
        if value is None:
            return ""
        ret: str = str(value)
        if self.delimiter in ret:
            ret = "\"" + ret + "\""
        return ret


class RowBundle:
    """
    A RowBundle represents a collection of information which may be used to print a row in a file.
    """
    index: int
    record: SapioRecord
    records: dict[str, SapioRecord]
    fields: dict[str, Any]
    may_skip: bool

    def __init__(self, index: int | None = None,
                 record: SapioRecord | None = None,
                 records: dict[str, SapioRecord] | None = None,
                 fields: dict[str, Any] | None = None,
                 may_skip: bool | None = None):
        """
        :param index: An index for this RowBundle. RowBundles are sorted by index before they are printed by the
            FileWriter. The FileWriter.build_file's sorter parameter is run after this sorting.
        :param record: A singular record for column definitions to pull information from.
        :param records: A dictionary of records for column definitions to pull information from. Each record is keyed by
            some name that column definitions can use to determine which record to get certain information from.
        :param fields: A list of "fields" specific to this bundle which aren't tied to a record.
        :param may_skip: If true, this RowBundle will return an empty string for ColumnDefs where may_skip is true.
        """
        self.index = index if index is not None else 0
        self.record = record
        self.records = records if records is not None else {}
        self.fields = fields if fields is not None else {}
        self.may_skip = may_skip if may_skip is not None else bool(records)


class ColumnDef:
    """
    The base class for all column definitions. Each column definition may cause a RowBundle to print a value in a
    different manner.
    """
    may_skip: bool
    """If true, this ColumnDef will return an empty string for RowBundles where may_skip is true."""

    @abstractmethod
    def print(self, row: RowBundle) -> Any:
        """
        :param row: The RowBundle to print some information for.
        :return: The printed value for the given RowBundle.
        """
        pass


class StaticColumn(ColumnDef):
    """
    A static column will always print the same value regardless of the input RowBundle.
    """
    value: Any

    def __init__(self, value: Any, may_skip: bool = False):
        """
        :param value: The value to print for this column.
        :param may_skip: If true, this ColumnDef will return an empty string for RowBundles where may_skip is true.
        """
        self.value = value
        self.may_skip = may_skip

    def print(self, row: RowBundle) -> Any:
        return self.value


class EmptyColumn(StaticColumn):
    """
    An empty column is a static column that always prints an empty string.
    """
    def __init__(self):
        super().__init__("")


class FieldSearchOrder(Enum):
    """
    An enum that specifies the order in which fields should be searched for in the RowBundle for FieldColumns.
    """
    RECORD_FIRST = 0
    """First search the fields on the record, then search the fields in the bundle."""
    BUNDLE_FIRST = 1
    """First search the fields in the bundle, then search the fields on the record."""
    RECORD_ONLY = 2
    """Only search the fields on the record."""
    BUNDLE_ONLY = 3
    """Only search the fields in the bundle."""


class FieldColumn(ColumnDef):
    """
    A field column prints the value of a given field from the input RowBundle. This field may come from a record in the
    RowBundle or from the RowBundle itself.
    """
    field_name: str
    record_key: str | None
    search_order: FieldSearchOrder
    skip_none_values: bool

    def __init__(self, field_name: str, record_key: str | None = None,
                 search_order: FieldSearchOrder = FieldSearchOrder.RECORD_FIRST,
                 skip_none_values: bool = False,
                 may_skip: bool = False):
        """
        :param field_name: The name of the field in the RowBundle to get the value of.
        :param record_key: If a record key is given, looks in the RowBundle's record dict for the record with they key.
            If no record key is given, look as the RowBundle's singular record.
        :param search_order: An enum that specifies the order in which fields should be searched for in the RowBundle.
        :param skip_none_values: If true and search_order is RECORD_FIRST or BUNDLE_FIRST, use the value of the second
            location's field if the first location's value is None.
        :param may_skip: If true, this ColumnDef will return an empty string for RowBundles where may_skip is true.
        """
        self.field_name = field_name
        self.record_key = record_key
        self.search_order = search_order
        self.skip_none_values = skip_none_values
        self.may_skip = may_skip

    def print(self, row: RowBundle) -> Any:
        return self._get_field(row)
    
    def _get_field(self, row: RowBundle) -> Any:
        record: SapioRecord = row.records.get(self.record_key) if self.record_key else row.record
        if self.search_order == FieldSearchOrder.RECORD_ONLY:
            return record.get_field_value(self.field_name) if record else None
        elif self.search_order == FieldSearchOrder.BUNDLE_ONLY:
            return row.fields.get(self.field_name)
        elif self.search_order == FieldSearchOrder.RECORD_FIRST:
            fields: dict[str, Any] = AliasUtil.to_field_map_lists([record])[0] if record else {}
            if self.field_name not in fields or (self.skip_none_values and fields.get(self.field_name) is None):
                return row.fields.get(self.field_name)
            return fields.get(self.field_name)
        elif self.search_order == FieldSearchOrder.BUNDLE_FIRST:
            if self.field_name not in row.fields or (self.skip_none_values and row.fields.get(self.field_name) is None):
                return record.get_field_value(self.field_name) if record else None
            return row.fields.get(self.field_name)


class DateColumn(FieldColumn):
    """
    A time column takes a field value which is an integer timestamp since the epoch and converts it to a human-readable
    date/time string.
    """
    time_format: str
    timezone: str | None = None

    def __init__(self, field_name: str, time_format: str, timezone: str | None = None, record_key: str | None = None,
                 search_order: FieldSearchOrder = FieldSearchOrder.RECORD_FIRST,
                 skip_none_values: bool = False, may_skip: bool = False):
        """
        :param field_name: The name of the field in the RowBundle to get the value of.
        :param time_format: The format to put the date in. See TimeUtil for more details on date formats.
        :param timezone: The timezone to convert the date to. If not specified, uses the default timezone of
            TimeUtil. See TimeUtil for more details on timezones.
        :param record_key: If a record key is given, looks in the RowBundle's record dict for the record with they key.
            If no record key is given, look as the RowBundle's singular record.
        :param search_order: An enum that specifies the order in which fields should be searched for in the RowBundle.
        :param skip_none_values: If true and search_order is RECORD_FIRST or BUNDLE_FIRST, use the value of the second
            location's field if the first location's value is None.
        :param may_skip: If true, this ColumnDef will return an empty string for RowBundles where may_skip is true.
        """
        super().__init__(field_name, record_key, search_order, skip_none_values, may_skip)
        self.time_format = time_format
        self.timezone = timezone

    def print(self, row: RowBundle) -> str:
        field: int = self._get_field(row)
        return TimeUtil.millis_to_format(field, self.time_format, self.timezone)


class ListColumn(ColumnDef):
    """
    A list column is like a static column, except it may print a different value for each row. If there are more rows
    than the length of this list, then all subsequent rows will print a blank value.
    
    Note that using the FileWriter.build_file's sorter parameter may result in these values not being in the expected
    order in the output file.
    """
    column_list: list[Any]
    list_size: int
    index: int = 0

    def __init__(self, column_list: list[Any]):
        """
        :param column_list: The list of values to print for this column.
        """
        self.column_list = column_list
        self.list_size = len(column_list)
        self.may_skip = False

    def print(self, row: RowBundle) -> Any:
        if self.index >= self.list_size:
            return ""
        val: Any = self.column_list[self.index]
        self.index += 1
        return val



from sapiopylib.rest.User import SapioUser

from sapiopycommons.general.aliases import RecordIdentifier, AliasUtil, ExperimentIdentifier, DataTypeIdentifier, \
    UserIdentifier


class ProcessTracking:
    @staticmethod
    def assign_to_process(context: UserIdentifier, data_type: DataTypeIdentifier, records: list[RecordIdentifier],
                          process_name: str, step_number: int | None = None, branch_id: int | None = None,
                          request: RecordIdentifier | None = None) -> None:
        """
        Assign the given tracked records to a new process at the specified step, settings its status to "Ready for -"
        at that step.
        Synonymous with what occurs during request creation or when using the assign to process button.

        :param context: The current webhook context or a user object to send requests from.
        :param data_type: The data type of the tracked records.
        :param records: A list of the tracked records.
        :param process_name: The name of the process that the tracked records should start at.
        :param step_number: The step number that the tracked records should start at. If not provided, assumes step
            number 1.
        :param branch_id: The branch ID of the above step. Only necessary to provide if multiple steps in the process
            share the same step number.
        :param request: The request that the tracked records are a part of, to be used for the assigned process record's
            request record ID field. If none is provided, creates a new request with default fields.
        """
        sub_path = '/ext/process-tracking/assign-to-process'
        payload = {
            "data-type-name": AliasUtil.to_data_type_name(data_type),
            "record-ids": AliasUtil.to_record_ids(records),
            "process-name": process_name,
            "step-number": step_number,
            "branch-id": branch_id,
            "request-record-id": AliasUtil.to_record_ids([request])[0] if request is not None else None
        }
        user: SapioUser = AliasUtil.to_sapio_user(context)
        response = user.post(sub_path, payload=payload)
        user.raise_for_status(response)

    @staticmethod
    def begin_protocol(context: UserIdentifier, data_type: DataTypeIdentifier, records: list[RecordIdentifier],
                       experiment: ExperimentIdentifier) -> None:
        """
        Begin the assigned processes of the given tracked records as the given experiment. This sets the status of the
        tracked records from "Ready for -" to "In Process -" for their current step.
        Synonymous with what occurs when starting a process step in the system.

        :param context: The current webhook context or a user object to send requests from.
        :param data_type: The data type of the tracked records.
        :param records: A list of the tracked records.
        :param experiment: The experiment that the tracked records are beginning the process for. This must be the next
            step in the current process of the tracked records.
        """
        sub_path = '/ext/process-tracking/begin-protocol'
        payload = {
            "data-type-name": AliasUtil.to_data_type_name(data_type),
            "record-ids": AliasUtil.to_record_ids(records),
            "experiment-id": AliasUtil.to_notebook_id(experiment),
        }
        user: SapioUser = AliasUtil.to_sapio_user(context)
        response = user.post(sub_path, payload=payload)
        user.raise_for_status(response)

    @staticmethod
    def complete_protocol(context: UserIdentifier, data_type: DataTypeIdentifier, records: list[RecordIdentifier],
                          experiment: ExperimentIdentifier) -> None:
        """
        Complete the current step that the given tracked records are at given the experiment.
        Moves the status to "Ready for -" for the next step in the process, or "Completed -" if this was the last
        step.
        Moves the status to "Failed -" if the failed flag for the tracked record is true.
        Moves the assigned process down to the descendant sample(s) if both samples are provided in the records list.
        Synonymous with what occurs when completing an experiment in the system.

        :param context: The current webhook context or a user object to send requests from.
        :param data_type: The data type of the tracked records.
        :param records: A list of the tracked records.
        :param experiment: The experiment that the tracked records are currently in and completing.
        """
        sub_path = '/ext/process-tracking/complete-protocol'
        payload = {
            "data-type-name": AliasUtil.to_data_type_name(data_type),
            "record-ids": AliasUtil.to_record_ids(records),
            "experiment-id": AliasUtil.to_notebook_id(experiment),
        }
        user: SapioUser = AliasUtil.to_sapio_user(context)
        response = user.post(sub_path, payload=payload)
        user.raise_for_status(response)

    @staticmethod
    def fail(context: UserIdentifier, data_type: DataTypeIdentifier, records: list[RecordIdentifier],
             experiment: ExperimentIdentifier) -> None:
        """
        Fail the assigned processes of the given tracked records, changing their statuses to "Failed -". The tracked
        records must be "In Process -" for the given step.
        Synonymous with what occurs when failing a sample due to a QC failure in a process in the system.

        :param context: The current webhook context or a user object to send requests from.
        :param data_type: The data type of the tracked records.
        :param records: A list of the tracked records.
        :param experiment: The experiment that the tracked records are currently in.
        """
        sub_path = '/ext/process-tracking/fail'
        payload = {
            "data-type-name": AliasUtil.to_data_type_name(data_type),
            "record-ids": AliasUtil.to_record_ids(records),
            "experiment-id": AliasUtil.to_notebook_id(experiment),
        }
        user: SapioUser = AliasUtil.to_sapio_user(context)
        response = user.post(sub_path, payload=payload)
        user.raise_for_status(response)

    @staticmethod
    def promote_to_next_by_experiment(context: UserIdentifier, data_type: DataTypeIdentifier,
                                      records: list[RecordIdentifier], experiment: ExperimentIdentifier) -> None:
        """
        Promote the status of the given tracked records to the next status in their process using an experiment.
        If the tracked records currently have a status of "Ready for -", then providing an experiment of the template
        they are ready for will move their status to "In Process -" for the experiment.
        If the tracked records currently have a status of "In Process -", then providing their current experiment
        will move their status to "Ready for -" for the next step in the process, or "Completed -" if this was the
        last step.

        :param context: The current webhook context or a user object to send requests from.
        :param data_type: The data type of the tracked records.
        :param records: A list of the tracked records.
        :param experiment: The experiment that the tracked records are currently in.
        """
        sub_path = '/ext/process-tracking/promote-status-to-next'
        payload = {
            "data-type-name": AliasUtil.to_data_type_name(data_type),
            "record-ids": AliasUtil.to_record_ids(records),
            "experiment-id": AliasUtil.to_notebook_id(experiment),
        }
        user: SapioUser = AliasUtil.to_sapio_user(context)
        response = user.post(sub_path, payload=payload)
        user.raise_for_status(response)

    @staticmethod
    def promote_to_next_by_step(context: UserIdentifier, data_type: DataTypeIdentifier,
                                records: list[RecordIdentifier], process_name: str, step_number: int,
                                branch_id: int | None = None) -> None:
        """
        Promote the status of the given tracked records to the next status in their process using step info.
        If the tracked records currently have a status of "Ready for -", then providing the step info for their current
        step will move their status to "In Process -" for that step.
        If the tracked records currently have a status of "In Process -", then providing the step info for their current
        step will move their status to "Ready for -" for the next step in the process, or "Completed -" if this was the
        last step.

        :param context: The current webhook context or a user object to send requests from.
        :param data_type: The data type of the tracked records.
        :param records: A list of the tracked records.
        :param process_name: The name of the process that the tracked records are currently in.
        :param step_number: The step number that the tracked records are currently in.
        :param branch_id: The branch ID of the above step. Only necessary to provide if multiple steps in the process
            share the same step number.
        """
        sub_path = '/ext/process-tracking/promote-status-to-next'
        payload = {
            "data-type-name": AliasUtil.to_data_type_name(data_type),
            "record-ids": AliasUtil.to_record_ids(records),
            "current-process-status": {
                "process-name": process_name,
                "step-number": step_number,
                "branch-id": branch_id
            }
        }
        user: SapioUser = AliasUtil.to_sapio_user(context)
        response = user.post(sub_path, payload=payload)
        user.raise_for_status(response)

    @staticmethod
    def reprocess(context: UserIdentifier, records: list[RecordIdentifier]) -> None:
        """
        Reprocess tracked records to a previous step in their process. Reprocessing is controlled by ReturnPoint records
        which are children of the AssignedProcess on the tracked records. Creates a new AssignedProcess record for the
        effected tracked records. Any existing AssignedProcess records are untouched.
        Synonymous with what occurs when reprocessing records to a previous step due to QC failures in a process in the
        system.

        :param context: The current webhook context or a user object to send requests from.
        :param records: A list of ReturnPoint records to reprocess to.
        """
        sub_path = '/ext/process-tracking/reprocess'
        payload = {
            "record-ids": AliasUtil.to_record_ids(records)
        }
        user: SapioUser = AliasUtil.to_sapio_user(context)
        response = user.post(sub_path, payload=payload)
        user.raise_for_status(response)



import base64
import io
import urllib.parse
from typing import Any

from requests import Response
from sapiopylib.rest.User import SapioUser

from sapiopycommons.general.aliases import UserIdentifier, AliasUtil


# FR-47387: Add support for the metadata endpoints in FileBridge.
class FileBridgeMetadata:
    """
    Metadata for a file or directory in FileBridge.
    """
    file_name: str
    """The name of the file or directory."""
    is_file: bool
    """True if the metadata is for a file, False if it is for a directory."""
    is_directory: bool
    """True if the metadata is for a directory, False if it is for a file."""
    size: int
    """The size of the file in bytes. For directories, this value will always be zero."""
    creation_time: int
    """The time the file or directory was created, in milliseconds since the epoch."""
    last_accessed_time: int
    """The time the file or directory was last accessed, in milliseconds since the epoch."""
    last_modified_time: int
    """The time the file or directory was last modified, in milliseconds since the epoch."""

    def __init__(self, json_dict: dict[str, Any]):
        self.file_name = json_dict['fileName']
        self.is_file = json_dict['isFile']
        self.is_directory = json_dict['isDirectory']
        self.size = json_dict['size']
        self.creation_time = json_dict['creationTime']
        self.last_accessed_time = json_dict['lastAccessTime']
        self.last_modified_time = json_dict['lastModifiedTime']


# FR-46064 - Initial port of PyWebhookUtils to sapiopycommons.
class FileBridge:
    @staticmethod
    def read_file(context: UserIdentifier, bridge_name: str, file_path: str,
                  base64_decode: bool = True) -> bytes:
        """
        Read a file from FileBridge.

        :param context: The current webhook context or a user object to send requests from.
        :param bridge_name: The name of the bridge to use. This is the "connection name" in the
            file bridge configurations.
        :param file_path: The path to read the file from.
        :param base64_decode: If true, base64 decode the file. Files are by default base64 encoded when retrieved from
            FileBridge.
        :return: The bytes of the file.
        """
        sub_path = '/ext/filebridge/readFile'
        params = {
            'Filepath': f"bridge://{bridge_name}/{file_path}"
        }
        user: SapioUser = AliasUtil.to_sapio_user(context)
        response = user.get(sub_path, params)
        user.raise_for_status(response)

        ret_val = response.content
        if base64_decode:
            ret_val = base64.b64decode(response.content)
        return ret_val

    @staticmethod
    def write_file(context: UserIdentifier, bridge_name: str, file_path: str,
                   file_data: bytes | str) -> None:
        """
        Write a file to FileBridge.

        :param context: The current webhook context or a user object to send requests from.
        :param bridge_name: The name of the bridge to use. This is the "connection name" in the
            file bridge configurations.
        :param file_path: The path to write the file to. If a file already exists at the given path then the file is
            overwritten.
        :param file_data: A string or bytes of the file to be written.
        """
        sub_path = '/ext/filebridge/writeFile'
        params = {
            'Filepath': f"bridge://{bridge_name}/{file_path}"
        }
        user: SapioUser = AliasUtil.to_sapio_user(context)
        with io.BytesIO(file_data.encode() if isinstance(file_data, str) else file_data) as data_stream:
            response = user.post_data_stream(sub_path, params=params, data_stream=data_stream)
        user.raise_for_status(response)

    @staticmethod
    def list_directory(context: UserIdentifier, bridge_name: str,
                       file_path: str | None = "") -> list[str]:
        """
        List the contents of a FileBridge directory.

        :param context: The current webhook context or a user object to send requests from.
        :param bridge_name: The name of the bridge to use. This is the "connection name" in the
            file bridge configurations.
        :param file_path: The path to read the directory from.
        :return: A list of names of files and folders in the directory.
        """
        sub_path = '/ext/filebridge/listDirectory'
        params = {
            'Filepath': f"bridge://{bridge_name}/{file_path}"
        }
        user: SapioUser = AliasUtil.to_sapio_user(context)
        response: Response = user.get(sub_path, params=params)
        user.raise_for_status(response)

        response_body: list[str] = response.json()
        path_length = len(f"bridge://{bridge_name}/")
        return [urllib.parse.unquote(value)[path_length:] for value in response_body]

    @staticmethod
    def create_directory(context: UserIdentifier, bridge_name: str, file_path: str) -> None:
        """
        Create a new directory in FileBridge.

        :param context: The current webhook context or a user object to send requests from.
        :param bridge_name: The name of the bridge to use. This is the "connection name" in the
            file bridge configurations.
        :param file_path: The path to create the directory at. If a directory already exists at the given path then an
            exception is raised.
        """
        sub_path = '/ext/filebridge/createDirectory'
        params = {
            'Filepath': f"bridge://{bridge_name}/{file_path}"
        }
        user: SapioUser = AliasUtil.to_sapio_user(context)
        response = user.post(sub_path, params=params)
        user.raise_for_status(response)

    @staticmethod
    def delete_file(context: UserIdentifier, bridge_name: str, file_path: str) -> None:
        """
        Delete an existing file in FileBridge.

        :param context: The current webhook context or a user object to send requests from.
        :param bridge_name: The name of the bridge to use. This is the "connection name" in the
            file bridge configurations.
        :param file_path: The path to the file to delete.
        """
        sub_path = '/ext/filebridge/deleteFile'
        params = {
            'Filepath': f"bridge://{bridge_name}/{file_path}"
        }
        user: SapioUser = AliasUtil.to_sapio_user(context)
        response = user.delete(sub_path, params=params)
        user.raise_for_status(response)

    @staticmethod
    def delete_directory(context: UserIdentifier, bridge_name: str, file_path: str) -> None:
        """
        Delete an existing directory in FileBridge.

        :param context: The current webhook context or a user object to send requests from.
        :param bridge_name: The name of the bridge to use. This is the "connection name" in the
            file bridge configurations.
        :param file_path: The path to the directory to delete.
        """
        sub_path = '/ext/filebridge/deleteDirectory'
        params = {
            'Filepath': f"bridge://{bridge_name}/{file_path}"
        }
        user: SapioUser = AliasUtil.to_sapio_user(context)
        response = user.delete(sub_path, params=params)
        user.raise_for_status(response)

    @staticmethod
    def file_metadata(context: UserIdentifier, bridge_name: str, file_path: str) -> FileBridgeMetadata:
        """
        Get metadata for a file or directory in FileBridge.

        The file path may be to a directory, in which case only the metadata for that directory will be returned. If you
        want the metadata for the contents of a directory, then use the directory_metadata function.

        :param context: The current webhook context or a user object to send requests from.
        :param bridge_name: The name of the bridge to use. This is the "connection name" in the
            file bridge configurations.
        :param file_path: The path to the file to retrieve the metadata from.
        :return: The metadata for the file.
        """
        sub_path = '/ext/filebridge/file/metadata'
        params = {
            'Filepath': f"bridge://{bridge_name}/{file_path}"
        }
        user: SapioUser = AliasUtil.to_sapio_user(context)
        response = user.get(sub_path, params=params)
        user.raise_for_status(response)
        response_body: dict[str, Any] = response.json()
        return FileBridgeMetadata(response_body)

    @staticmethod
    def directory_metadata(context: UserIdentifier, bridge_name: str, file_path: str) -> list[FileBridgeMetadata]:
        """
        Get metadata for every file or nested directory in a directory in FileBridge.

        :param context: The current webhook context or a user object to send requests from.
        :param bridge_name: The name of the bridge to use. This is the "connection name" in the
            file bridge configurations.
        :param file_path: The path to the directory to retrieve the metadata of the contents.
        :return: A list of the metadata for the contents of the directory.
        """
        sub_path = '/ext/filebridge/directory/metadata'
        params = {
            'Filepath': f"bridge://{bridge_name}/{file_path}"
        }
        user: SapioUser = AliasUtil.to_sapio_user(context)
        response = user.get(sub_path, params=params)
        user.raise_for_status(response)
        response_body: list[dict[str, Any]] = response.json()
        return [FileBridgeMetadata(x) for x in response_body]



from __future__ import annotations

import re
from abc import abstractmethod
from typing import Any

from sapiopylib.rest.User import SapioUser
from sapiopylib.rest.pojo.CustomReport import RawReportTerm, RawTermOperation
from sapiopylib.rest.pojo.datatype.FieldDefinition import VeloxIntegerFieldDefinition, VeloxStringFieldDefinition, \
    AbstractVeloxFieldDefinition

from sapiopycommons.callbacks.callback_util import CallbackUtil
from sapiopycommons.files.file_data_handler import FileDataHandler, FilterList
from sapiopycommons.general.aliases import UserIdentifier, AliasUtil
from sapiopycommons.general.custom_report_util import CustomReportUtil
from sapiopycommons.general.exceptions import SapioUserCancelledException
from sapiopycommons.general.time_util import TimeUtil


class FileValidator:
    """
    The FileValidator is a class used to verify that a file provided by a user meets certain specs. For example, you
    may require that all cells in a column not be blank, or have a certain range of values, or that no two values in
    a column are the same. These validation rules can be added to the FileValidator along with the data for the file.
    Then, when the file is validated, the index of any rows that failed a rule will be returned along with the list of
    rules that they failed, allowing you to report the specifics of why a rule failed to the user.

    Look into using this in combination with FileDataHandler to prepare files for the FileValidator and for use in
    data record fields.
    """
    file_data: list[dict[str, Any]]
    rules: list[ValidationRule]

    def __init__(self, file_data: list[dict[str, Any]]):
        """
        :param file_data: A list of dictionaries. Every dictionary in the list is expected to have the same keys.
            FileUtil.tokenize_csv and tokenize_xlsx can be used to convert a file into such a list.
            CustomReportUtil can also generate lists of dictionaries that match this criteria.
        """
        self.file_data = file_data
        self.rules = []

    def add_rule(self, rule: ValidationRule) -> None:
        """
        Add a new validation rule to this validator. When the validate_file function is called, each added rule will be
        run on the file. Rules are evaluated in the same order that they are added to the validator.

        Custom validation rules can be created by defining a class that extends RowRule or ColumnRule and implements
        the validation method.

        :param rule: A validation rule to be run when the file is validated.
        """
        self.rules.append(rule)

    def validate_file(self) -> dict[int, list[ValidationRule]]:
        """
        Validate the file, returning any rule failures that are encountered.

        :return: A dictionary mapping row indices to a list of the rules that they failed. This can then be used to
            report errors back to the user who uploaded the file by checking the class type of the rules.
        """
        failed_rows: dict[int, list[ValidationRule]] = {}

        # Check each rule for this validator.
        for rule in self.rules:
            # If this is a row rule, then the rule only runs on a per-row basis. Iterate over every row in the
            # file and use the rule to validate them.
            if isinstance(rule, RowRule):
                for index, row in enumerate(self.file_data):
                    if FileDataHandler.skip_row(index, row, whitelist=rule.whitelist, blacklist=rule.blacklist):
                        continue
                    # These rules return a boolean for whether the rule passed or not.
                    if not rule.validate(row):
                        failed_rows.setdefault(index, []).append(rule)
            # If this is a column rule, then the rule runs down an entire column in the file. Pass the entire file data
            # list to the rule for validation.
            elif isinstance(rule, ColumnRule):
                # These rules return a list of row indices that caused the rule to fail for the entire column.
                for index in rule.validate(self.file_data):
                    failed_rows.setdefault(index, []).append(rule)

        return failed_rows

    def build_violation_report(self, context: UserIdentifier,
                               rule_violations: dict[int, list[ValidationRule]]) -> None:
        """
        Display a simple report of any rule violations in the file to the user as a table dialog.

        :param context: The current webhook context or a user object to send requests from.
        :param rule_violations: A dict of rule violations generated by a call to validate_file.
        """
        if not rule_violations:
            return

        file_handler = FileDataHandler(self.file_data)
        columns: list[AbstractVeloxFieldDefinition] = [
            VeloxIntegerFieldDefinition("Errors", "RowNum", "Row Number"),
            VeloxStringFieldDefinition("Errors", "Header", "Header"),
            VeloxStringFieldDefinition("Errors", "Value", "Value", max_length=500),
            VeloxStringFieldDefinition("Errors", "Reason", "Reason", max_length=2000, default_table_column_width=500)
        ]
        rows: list[dict[str, Any]] = []
        for index, violations in rule_violations.items():
            file_row: dict[str, Any] = file_handler.get_row(index)
            for violation in violations:
                if isinstance(violation, ColumnRule):
                    # Column rules always act upon a specific header, so list the header and the value for this run
                    # under that header alongside the reason.
                    rows.append({
                        "RowNum": index + 2,
                        "Header": violation.header,
                        "Value": str(file_row.get(violation.header)),
                        "Reason": violation.reason[:2000]
                    })
                elif isinstance(violation, RowRule):
                    # We can't know what header(s) a row rule was looking at, so just leave header and value
                    # blank and trust that the reason describes enough what the issue was.
                    rows.append({
                        "RowNum": index + 2,
                        "Header": "N/A",
                        "Value": "N/A",
                        "Reason": violation.reason[:2000]
                    })

        callback = CallbackUtil(context)
        callback.table_dialog("Errors", "The following rule violations were encountered in the provided file.",
                              columns, rows)

    def validate_and_report_errors(self, context: UserIdentifier) -> None:
        """
        Validate the file. If any rule violations are found, display a simple report of any rule violations in the file
        to the user as a table dialog and throw a SapioUserCancelled exception after the user acknowledges the dialog
        to end the webhook interaction.

        Shorthand for calling validate_file() and then build_violation_report() if there are any errors.

        :param context: The current webhook context or a user object to send requests from.
        """
        violations = self.validate_file()
        if violations:
            self.build_violation_report(context, violations)
            raise SapioUserCancelledException()


class ValidationRule:
    """
    The base class for all rules. Each rule has a reason that can be provided for why a file violated the rule and
    optional whitelists and blacklists for rows to skip checking the rule for.

    This class should not be extended. If you want to create a custom rule, extend RowRule or ColumnRule.
    """
    reason: str
    whitelist: FilterList
    blacklist: FilterList

    def __init__(self, reason: str, whitelist: FilterList, blacklist: FilterList):
        """
        :param reason: A string explaining the reason why an violation occurred for this rule.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        """
        self.reason = reason
        self.blacklist = blacklist
        self.whitelist = whitelist


class RowRule(ValidationRule):
    """
    The base class for all row rules. Row rules aren't tied to a specific header and act across multiple cells in a row.
    Extend this class to create your own custom file validation rules.
    """
    def __init__(self, reason: str, whitelist: FilterList = None, blacklist: FilterList = None):
        """
        :param reason: A string explaining the reason why an violation occurred for this rule.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        """
        super().__init__(reason, whitelist, blacklist)

    @abstractmethod
    def validate(self, row: dict[str, Any]) -> bool:
        """
        Validate that a row meets the requirements of this rule.

        :param row: A row from the file.
        :return: Whether the given row has passed the rule validation.
        """
        pass


class ColumnRule(ValidationRule):
    """
    The base class for all column rules. Column rules are tied to specific a header and act upon every cell under that
    header. Extend this class to create your own custom file validation rules.
    """
    header: str

    def __init__(self, header: str, reason: str, whitelist: FilterList = None, blacklist: FilterList = None):
        """
        :param header: The header that this rule acts upon.
        :param reason: A string explaining the reason why a violation occurred for this rule.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        """
        super().__init__(reason, whitelist, blacklist)
        self.header = header

    @abstractmethod
    def validate(self, rows: list[dict[str, Any]]) -> list[int]:
        """
        Validate that a column meets the requirements of this rule.

        :param rows: Every row from the file.
        :return: A list of the indices of all rows that failed the rule validation.
        """
        pass


class NoBlanksRule(ColumnRule):
    """
    Require that every cell in a column have a non-blank value.
    """
    def __init__(self, header: str, *, reason: str | None = None, whitelist: FilterList = None,
                 blacklist: FilterList = None):
        """
        :param header: The header that this rule acts upon.
        :param reason: A string explaining the reason why a violation occurred for this rule. A default reason is
            provided if none is given by this initialization, but you may want to provide more detailed reasoning than
            the default.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        """
        if reason is None:
            reason = "This value is not allowed to be blank."
        super().__init__(header, reason, whitelist, blacklist)

    def validate(self, rows: list[dict[str, Any]]) -> list[int]:
        handler = FileDataHandler(rows)
        return handler.empty_cells(self.header, whitelist=self.whitelist, blacklist=self.blacklist)


class NoDuplicatesRule(ColumnRule):
    """
    Require that no two values in a column be duplicates.
    """
    def __init__(self, header: str, *, reason: str | None = None, whitelist: FilterList = None,
                 blacklist: FilterList = None):
        """
        :param header: The header that this rule acts upon.
        :param reason: A string explaining the reason why a violation occurred for this rule. A default reason is
            provided if none is given by this initialization, but you may want to provide more detailed reasoning than
            the default.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        """
        if reason is None:
            reason = "This value is a duplicate of another value in the same column."
        super().__init__(header, reason, whitelist, blacklist)

    def validate(self, rows: list[dict[str, Any]]) -> list[int]:
        handler = FileDataHandler(rows)
        duplicates: dict[Any, list[int]] = handler.get_duplicates(self.header,
                                                                  whitelist=self.whitelist, blacklist=self.blacklist)
        indices: list[int] = []
        for cells in duplicates.values():
            indices += cells
        return indices


class AllowedValuesRule(ColumnRule):
    """
    Require that every cell in a column has a value that is present in the given list.
    """
    values: list[Any]

    def __init__(self, header: str, values: list[Any], *, reason: str | None = None, whitelist: FilterList = None,
                 blacklist: FilterList = None):
        """
        :param header: The header that this rule acts upon.
        :param values: A list of any values.
        :param reason: A string explaining the reason why a violation occurred for this rule. A default reason is
            provided if none is given by this initialization, but you may want to provide more detailed reasoning than
            the default.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        """
        if reason is None:
            reason = f"This value is not one of the allowed values: {', '.join([str(x) for x in values])}"
        super().__init__(header, reason, whitelist, blacklist)
        self.values = values

    def validate(self, rows: list[dict[str, Any]]) -> list[int]:
        handler = FileDataHandler(rows)
        return handler.get_not_in_list(self.header, self.values, whitelist=self.whitelist, blacklist=self.blacklist)


class MatchesPatternRule(ColumnRule):
    """
    Requires that every cell in a column has a value that matches a regex pattern.
    """
    pattern: str

    def __init__(self, header: str, pattern: str | re.Pattern[str], *, reason: str | None = None,
                 whitelist: FilterList = None, blacklist: FilterList = None):
        """
        :param header: The header that this rule acts upon.
        :param pattern: A regex pattern.
        :param reason: A string explaining the reason why a violation occurred for this rule. A default reason is
            provided if none is given by this initialization, but you may want to provide more detailed reasoning than
            the default.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        """
        if reason is None:
            reason = f"This value does not match the expected format: {pattern}."
        super().__init__(header, reason, whitelist, blacklist)
        self.pattern = pattern

    def validate(self, rows: list[dict[str, Any]]) -> list[int]:
        handler = FileDataHandler(rows)
        return handler.get_mismatches(self.header, self.pattern, whitelist=self.whitelist, blacklist=self.blacklist)


class MatchesDateFormatRule(ColumnRule):
    """
    Requires that every cell in a column has a value that matches a given date format.
    """
    time_format: str

    def __init__(self, header: str, time_format: str, *, reason: str | None = None, whitelist: FilterList = None,
                 blacklist: FilterList = None):
        """
        :param header: The header that this rule acts upon.
        :param time_format: A date/time format. See TimeUtil for more specifics.
        :param reason: A string explaining the reason why a violation occurred for this rule. A default reason is
            provided if none is given by this initialization, but you may want to provide more detailed reasoning than
            the default.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        """
        if reason is None:
            reason = f"This value does not match the expected format: {time_format}."
        super().__init__(header, reason, whitelist, blacklist)
        self.time_format = time_format

    def validate(self, rows: list[dict[str, Any]]) -> list[int]:
        handler = FileDataHandler(rows)
        return handler.get_by_function(lambda i, r: not TimeUtil.str_matches_format(r.get(self.header), self.time_format),
                                       whitelist=self.whitelist, blacklist=self.blacklist)


class ValueIsCastableRule(ColumnRule):
    """
    Requires that every cell in a column has a value that can be cast to a given type. Intended for determining if
    string values can be cast to ints or floats, but you may provide other types as well. Works by checking if
    type(value) raises an exception, so if the type you provide wouldn't raise an exception for a bad string, then
    this rule will never fail.
    """
    cast_type: type

    def __init__(self, header: str, cast_type: type, *, reason: str | None = None, whitelist: FilterList = None,
                 blacklist: FilterList = None):
        """
        :param header: The header that this rule acts upon.
        :param cast_type: A type to attempt to cast the values under the header to.
        :param reason: A string explaining the reason why a violation occurred for this rule. A default reason is
            provided if none is given by this initialization, but you may want to provide more detailed reasoning than
            the default.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        """
        if reason is None:
            reason = f"This value cannot be casted to the type {cast_type}"
        super().__init__(header, reason, whitelist, blacklist)
        self.cast_type = cast_type

    def validate(self, rows: list[dict[str, Any]]) -> list[int]:
        # Get the index of every row that can't be cast to the desired type.
        def func(index: int, row: dict[str, Any]) -> bool:
            try:
                self.cast_type(row.get(self.header))
                return False
            except Exception:
                return True

        handler = FileDataHandler(rows)
        return handler.get_by_function(func, whitelist=self.whitelist, blacklist=self.blacklist)


class ValueInsideRangeRule(ColumnRule):
    """
    Requires that every cell in a column has a value that is inside the given range (inclusive).
    """
    min_val: float | int
    max_val: float | int

    def __init__(self, header: str, min_val: float | int, max_val: float | int,
                 *, reason: str | None = None, whitelist: FilterList = None, blacklist: FilterList = None):
        """
        :param header: The header that this rule acts upon.
        :param min_val: The minimum allowed value of the cell.
        :param max_val: The maximum allowed value of the cell.
        :param reason: A string explaining the reason why a violation occurred for this rule. A default reason is
            provided if none is given by this initialization, but you may want to provide more detailed reasoning than
            the default.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        """
        if reason is None:
            reason = f"This value is not within the range [{min_val}, {max_val}]."
        super().__init__(header, reason, whitelist, blacklist)
        self.min_val = min_val
        self.max_val = max_val

    def validate(self, rows: list[dict[str, Any]]) -> list[int]:
        handler = FileDataHandler(rows)
        return handler.get_inside_range(self.header, self.min_val, self.max_val,
                                        whitelist=self.whitelist, blacklist=self.blacklist)


class ValueOutsideRangeRule(ColumnRule):
    """
    Requires that every cell in a column has a value that is outside the given range (exclusive).
    """
    min_val: float | int
    max_val: float | int

    def __init__(self, header: str, min_val: float | int, max_val: float | int,
                 *, reason: str | None = None, whitelist: FilterList = None, blacklist: FilterList = None):
        """
        :param header: The header that this rule acts upon.
        :param min_val: The value that the cell may be below.
        :param max_val: The value that the cell may be above.
        :param reason: A string explaining the reason why a violation occurred for this rule. A default reason is
            provided if none is given by this initialization, but you may want to provide more detailed reasoning than
            the default.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        """
        if reason is None:
            reason = f"This value is not outside of the range [{min_val}, {max_val}]"
        super().__init__(header, reason, whitelist, blacklist)
        self.min_val = min_val
        self.max_val = max_val

    def validate(self, rows: list[dict[str, Any]]) -> list[int]:
        handler = FileDataHandler(rows)
        return handler.get_outside_range(self.header, self.min_val, self.max_val,
                                         whitelist=self.whitelist, blacklist=self.blacklist)


class ContainsSubstringFromCellRule(RowRule):
    """
    Requires that the cell in the first column contains the value in the cell of the second column.
    """
    first: str
    second: str

    def __init__(self, first: str, second: str, *, reason: str | None = None, whitelist: FilterList = None,
                 blacklist: FilterList = None):
        """
        :param first: The header to check the contents up.
        :param second: The header to use the values of to check the above header.
        :param reason: A string explaining the reason why a violation occurred for this rule. A default reason is
            provided if none is given by this initialization, but you may want to provide more detailed reasoning than
            the default.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        """
        if reason is None:
            reason = f"The value in column {first} does not contain the value in the column {second}."
        super().__init__(reason, whitelist, blacklist)
        self.first = first
        self.second = second

    def validate(self, row: dict[str, Any]) -> bool:
        return row.get(self.second) in row.get(self.first)


class UniqueSystemValueRule(ColumnRule):
    """
    Requires that every cell in the column has a value that is not already in use in the system for a given data type
    and field name.
    """
    user: SapioUser
    data_type_name: str
    data_field_name: str

    def __init__(self, context: UserIdentifier, header: str, data_type_name: str,
                 data_field_name: str):
        """
        :param context: The current webhook context or a user object to send requests from.
        :param header: The header that this rule acts upon.
        :param data_type_name: The data type name to search on.
        :param data_field_name: The data field name to search on. This is expected to be a string field.
        """
        self.user = AliasUtil.to_sapio_user(context)
        self.data_type_name = data_type_name
        self.data_field_name = data_field_name
        super().__init__(header, f"This value already exists in the system.")

    def validate(self, rows: list[dict[str, Any]]) -> list[int]:
        file_handler = FileDataHandler(rows)
        values: list[str] = file_handler.get_values_list(self.header)

        # Run a quick report for all records of this type that match these field values.
        term = RawReportTerm(self.data_type_name, self.data_field_name, RawTermOperation.EQUAL_TO_OPERATOR,
                             "{" + ",".join(values) + "}")
        results: list[dict[str, Any]] = CustomReportUtil.run_quick_report(self.user, term)
        existing_values: list[Any] = [x.get(self.data_field_name) for x in results]
        return file_handler.get_in_list(self.header, existing_values)


class ExistingSystemValueRule(ColumnRule):
    """
    Requires that every cell in the column has a value that is already in use in the system for a given data type
    and field name.
    """
    user: SapioUser
    data_type_name: str
    data_field_name: str

    def __init__(self, context: UserIdentifier, header: str, data_type_name: str,
                 data_field_name: str):
        """
        :param context: The current webhook context or a user object to send requests from.
        :param header: The header that this rule acts upon.
        :param data_type_name: The data type name to search on.
        :param data_field_name: The data field name to search on. This is expected to be a string field.
        """
        self.user = AliasUtil.to_sapio_user(context)
        self.data_type_name = data_type_name
        self.data_field_name = data_field_name
        super().__init__(header, f"This value doesn't exist in the system.")

    def validate(self, rows: list[dict[str, Any]]) -> list[int]:
        file_handler = FileDataHandler(rows)
        values: list[str] = file_handler.get_values_list(self.header)

        # Run a quick report for all records of this type that match these field values.
        term = RawReportTerm(self.data_type_name, self.data_field_name, RawTermOperation.EQUAL_TO_OPERATOR,
                             "{" + ",".join(values) + "}")
        results: list[dict[str, Any]] = CustomReportUtil.run_quick_report(self.user, term)
        existing_values: list[Any] = [x.get(self.data_field_name) for x in results]
        return file_handler.get_not_in_list(self.header, existing_values)
